
/****************************************************************************
 * This file is part of the program  PhyloQFold 			                *
 * for MCMC-sampling consensus secondary structures with pseudoknot         *
 * for a given alignment of RNA sequences and a given phylogeny.            *                                       
 *                                                                          *
 *    Copyright (C) 2012  Dirk Metzler & Gero Doose                        *
 *    e-mail: metzler@bio.lmu.de and gero.doose@gmx.de                      *
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/


#ifndef EVOMOD_H_
#define EVOMOD_H_
#include "Matrix.h"
#include "Node.h"
#include <iostream>
#include <fstream>
#include <cmath>
#include <cstdlib>




class EvoMod {
private:
	void loadMatrices(vector<double>* const  & mup, Matrix * const  & mpp,  Matrix * const & mur,  Matrix * const & mpr);
	void matrixMulti(Matrix * const & multi, Matrix * const & ma, long double  d);
	void matrixSquare(Matrix  * const & square);
	void matrixAdition(Matrix * const & ma, Matrix * const & mb);
	void transitionDynamicsPair(Matrix * const & pt,Matrix * const & pairedR, double const dist);
	void transitionDynamics(Matrix * const & pt,  Matrix * const & unpairedR, double const dist);
	void calculateTransitionMatricesReal(Node * const & r, Matrix * const & unpairedR, Matrix * const & pairedR);
	void calculateTransitionMatrices(Node * const & r, Matrix * const & unpairedR, Matrix * const & pairedR);
public:
	EvoMod(Node * const &, vector<double>* const  & mup, Matrix * const  & mpp,  Matrix * const & mur,  Matrix * const & mpr);
	virtual ~EvoMod();
};

#endif /* EVOMOD_H_ */
