

/****************************************************************************
 * This file is part of the program  PhyloQFold 			                *
 * for MCMC-sampling consensus secondary structures with pseudoknot         *
 * for a given alignment of RNA sequences and a given phylogeny.            *                                       
 *                                                                          *
 *    Copyright (C) 2012  Dirk Metzler & Gero Doose                        *
 *    e-mail: metzler@bio.lmu.de and gero.doose@gmx.de                      *
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#include "Likelihood.h"


Likelihood::Likelihood(map<string, double> * const & columnLikelihood,map<string,
		double> * const & columnPairLikelihood,Node * const & n,vector<string> * const alignment,
		vector<double> * const & m_unpairedProbs,
		Matrix * const & m_pairedProbs,const char * const dateiname)
{
	readEpsilon(dateiname);
	calculateLikelihood(columnLikelihood,columnPairLikelihood,n,alignment, m_unpairedProbs, m_pairedProbs );
}

Likelihood::~Likelihood() {

}

void Likelihood::readEpsilon(const char * const dateiname){
    ifstream f(dateiname);
	string s;
	f >> s;
	if(s!= "Epsilon"){
		cerr<<"error in file inputgrammer"<<endl;
		exit(1);
	}
	else{
		f >> epsilon;
	}
}

void Likelihood::resetprobK(Node * const &  r) {
   if (!r->isLeaf()) {
	   for(unsigned int i=0;i<4;i++){
		   r->getProbK().at(i)=-1;
	   }
	   for(unsigned int j=0;j<16;j++){
		   r->getProbKpair().at(j)=-1;
	   }
	   resetprobK( r->getLeftChild());
	   resetprobK( r->getRightChild());
   }
   else{
	   for(unsigned int i=0;i<4;i++){
		   r->getProbK().at(i)=-1;
	   }
	   for(unsigned int j=0;j<16;j++){
		   r->getProbKpair().at(j)=-1;
	   }
   }
}
int Likelihood::charToInt(char const a){
	int i;
	switch(a){
		case'a':
		case'A':
			i=0;
			break;
		case'c':
		case'C':
			i=1;
			break;
		case'g':
		case'G':
			i=2;
			break;
		case'u':
		case'U':
		case't':
		case'T':
			i=3;
			break;
	}
	return i;
}
char Likelihood::intToChar(int const a){
	char i;
	switch(a){
	case 0:
		i='a';
		break;
	case 1:
		i='c';
		break;
	case 2:
		i='g';
		break;
	case 3:
		i='u';
		break;

	}
	return i;
}
int Likelihood::charToIntPair(string const a){
	int i;
	switch(a.at(0)){
	case'a':
	case'A':
		switch(a.at(1)){
			case'a':
			case'A':
				i=0;
				break;
			case'c':
			case'C':
				i=1;
				break;
			case'g':
			case'G':
				i=2;
				break;
			case'u':
			case'U':
			case't':
			case'T':
				i=3;
				break;
		}
		break;
	case'c':
	case'C':
		switch(a.at(1)){
			case'a':
			case'A':
				i=4;
				break;
			case'c':
			case'C':
				i=5;
				break;
			case'g':
			case'G':
				i=6;
				break;
			case'u':
			case'U':
			case't':
			case'T':
				i=7;
				break;
		}
		break;
	case'g':
	case'G':
		switch(a.at(1)){
			case'a':
			case'A':
				i=8;
				break;
			case'c':
			case'C':
				i=9;
				break;
			case'g':
			case'G':
				i=10;
				break;
			case'u':
			case'U':
			case't':
			case'T':
				i=11;
				break;
		}
		break;
	case'u':
	case'U':
	case't':
	case'T':
		switch(a.at(1)){
			case'a':
			case'A':
				i=12;
				break;
			case'c':
			case'C':
				i=13;
				break;
			case'g':
			case'G':
				i=14;
				break;
			case'u':
			case'U':
			case'T':
			case't':
				i=15;
				break;
		}
		break;
	}
	return i;
}
string Likelihood::intToCharPair(int const a){
	string i;
	switch(a){
	case 0:
		i="aa";
		break;
	case 1:
		i="ac";
		break;
	case 2:
		i="ag";
		break;
	case 3:
		i="au";
		break;
	case 4:
		i="ca";
		break;
	case 5:
		i="cc";
		break;
	case 6:
		i="cg";
		break;
	case 7:
		i="cu";
		break;
	case 8:
		i="ga";
		break;
	case 9:
		i="gc";
		break;
	case 10:
		i="gg";
		break;
	case 11:
		i="gu";
		break;
	case 12:
		i="ua";
		break;
	case 13:
		i="uc";
		break;
	case 14:
		i="ug";
		break;
	case 15:
		i="uu";
		break;
	}
	return i;
}
 ////gap behandlung: cutten des gesammten astes!
double Likelihood::probKPairGapAmbiguity(Node * const & t, string  const & columnA, string const & columnB, unsigned int x ){
	if(t->isLeaf()){
		cerr<<"tree size error (tree consists of only one leaf)"<<endl;
		exit(1);
	}
	double termLeft =0;
	if(t->getLeftChild()->isLeaf()){
		//P(x->y) fuer EIN y
		char yCharOne =columnA.at(t->getLeftChild()->getAlignmentRaw());
		char yCharTwo = columnB.at(t->getLeftChild()->getAlignmentRaw());
		string yString="";
		yString.push_back(yCharOne);
		yString.push_back(yCharTwo);
		if(yCharOne !='-' && yCharTwo != '-'){
			unsigned int y=charToIntPair(yString);
			termLeft =t->getLeftChild()->getPtPaired()->getTable()[x][y];
		}
		else{
			if(yCharOne !='-' && yCharTwo == '-'){
				switch(charToInt(yCharOne)){
					case 0:
						for(unsigned int i=0;i<4;i++){
							double p=0;
							p=t->getLeftChild()->getPtPaired()->getTable()[x][i];
							termLeft += p;
						}
						break;
					case 1:
						for(unsigned int i=4;i<8;i++){
							double p=0;
							p=t->getLeftChild()->getPtPaired()->getTable()[x][i];
							termLeft += p;
						}
						break;
					case 2:
						for(unsigned int i=8;i<12;i++){
							double p=0;
							p=t->getLeftChild()->getPtPaired()->getTable()[x][i];
							termLeft += p;
						}
						break;
					case 3:
						for(unsigned int i=12;i<16;i++){
							double p=0;
							p=t->getLeftChild()->getPtPaired()->getTable()[x][i];
							termLeft += p;
						}
						break;
				}

			}
			else{
				if(yCharOne =='-' && yCharTwo != '-'){
					switch(charToInt(yCharTwo)){
						case 0:
							for(unsigned int i=0;i<16;i=i+4){
								double p=0;
								p=t->getLeftChild()->getPtPaired()->getTable()[x][i];
								termLeft += p;
							}
							break;
						case 1:
							for(unsigned int i=1;i<16;i=i+4){
								double p=0;
								p=t->getLeftChild()->getPtPaired()->getTable()[x][i];
								termLeft += p;
							}
							break;
						case 2:
							for(unsigned int i=2;i<16;i=i+4){
								double p=0;
								p=t->getLeftChild()->getPtPaired()->getTable()[x][i];
								termLeft += p;
							}
							break;
						case 3:
							for(unsigned int i=3;i<16;i=i+4){
								double p=0;
								p=t->getLeftChild()->getPtPaired()->getTable()[x][i];
								termLeft += p;
							}
							break;
					}
				}
				else{
					/*
					for(unsigned int i=0;i<16;i++){
						 double p=0;
						 p=t->getLeftChild()->getPtPaired()->getTable()[x][i];
						 termLeft += p;
					}
					*/
					termLeft=1;
				}
			}
		}

	}
	else{
		//rekursive summe ueber alle basenPaare y
		for(unsigned int i=0;i<16;i++){
			 double p=0;
			 p=t->getLeftChild()->getPtPaired()->getTable()[x][i];
			 if(t->getLeftChild()->getProbKpair().at(i)!=-1){
				 termLeft += p * t->getLeftChild()->getProbKpair().at(i);
			 }
			 else{
				 termLeft += p * probKPairGapAmbiguity(t->getLeftChild(), columnA,columnB, i);
			 }
		}
	}
	double termRight =0;
	if(t->getRightChild()->isLeaf()){
		char yCharOne =columnA.at(t->getRightChild()->getAlignmentRaw());
		char yCharTwo = columnB.at(t->getRightChild()->getAlignmentRaw());
		string yString="";
		yString.push_back(yCharOne);
		yString.push_back(yCharTwo);
		if(yCharOne !='-' && yCharTwo != '-'){
			unsigned int y=charToIntPair(yString);
			termRight =t->getRightChild()->getPtPaired()->getTable()[x][y];
		}
		else{
			if(yCharOne !='-' && yCharTwo == '-'){
				switch(charToInt(yCharOne)){
					case 0:
						for(unsigned int i=0;i<4;i++){
							double p=0;
							p=t->getRightChild()->getPtPaired()->getTable()[x][i];
							termRight += p;
						}
						break;
					case 1:
						for(unsigned int i=4;i<8;i++){
							double p=0;
							p=t->getRightChild()->getPtPaired()->getTable()[x][i];
							termRight += p;
						}
						break;
					case 2:
						for(unsigned int i=8;i<12;i++){
							double p=0;
							p=t->getRightChild()->getPtPaired()->getTable()[x][i];
							termRight += p;
						}
						break;
					case 3:
						for(unsigned int i=12;i<16;i++){
							double p=0;
							p=t->getRightChild()->getPtPaired()->getTable()[x][i];
							termRight += p;
						}
						break;
				}

			}
			else{
				if(yCharOne =='-' && yCharTwo != '-'){
					switch(charToInt(yCharTwo)){
						case 0:
							for(unsigned int i=0;i<16;i=i+4){
								double p=0;
								p=t->getRightChild()->getPtPaired()->getTable()[x][i];
								termRight += p;
							}
							break;
						case 1:
							for(unsigned int i=1;i<16;i=i+4){
								double p=0;
								p=t->getRightChild()->getPtPaired()->getTable()[x][i];
								termRight += p;
							}
							break;
						case 2:
							for(unsigned int i=2;i<16;i=i+4){
								double p=0;
								p=t->getRightChild()->getPtPaired()->getTable()[x][i];
								termRight += p;
							}
							break;
						case 3:
							for(unsigned int i=3;i<16;i=i+4){
								double p=0;
								p=t->getRightChild()->getPtPaired()->getTable()[x][i];
								termRight += p;
							}
							break;
					}
				}
				else{
					/*
					for(unsigned int i=0;i<16;i++){
						 double p=0;
						 p=t->getRightChild()->getPtPaired()->getTable()[x][i];
						 termRight += p;
					}
					*/
					termRight=1;
				}
			}
		}
	}else{
		for(unsigned int i=0;i<16;i++){
			double p=0;
			 p=t->getRightChild()->getPtPaired()->getTable()[x][i];
			 if(t->getRightChild()->getProbKpair().at(i)!=-1){
				 termRight += p * t->getRightChild()->getProbKpair().at(i);
			 }
			 else{
				 termRight += p * probKPairGapAmbiguity(t->getRightChild(), columnA,columnB, i);
			 }
		}
	}
	t->getProbKpair().at(x)=(termLeft * termRight);
	return (termLeft * termRight);
}


//gap behandlung: cutten des gesammten astes! ein ganzer term wird 1 dadurch geht nur der ast ein der ein nukleotid für diese spalte bereithaelt
double Likelihood::probKGapAmbiguity(Node * const & t, string const & column, unsigned int const x ){
	if(t->isLeaf()){
		cerr<<"tree size error (tree consists of only one leaf)"<<endl;
		exit(1);
	}
	double termLeft =0;
	if(t->getLeftChild()->isLeaf()){
		//P(x->y) (am rekursionsanker)
		char yChar= column.at(t->getLeftChild()->getAlignmentRaw());
		if(yChar=='-'){
			termLeft=1;
		}
		else{
			unsigned int y= charToInt(yChar);
			termLeft =t->getLeftChild()->getPt()->getTable()[x][y];
		}

	}else{
		//rekursion: summe ueber alle basen y
		for(unsigned int i=0;i<4;i++){
			 double p=0;
			 p=t->getLeftChild()->getPt()->getTable()[x][i];
			 if(t->getLeftChild()->getProbK().at(i)!=-1){
				 termLeft += p * t->getLeftChild()->getProbK().at(i);
			 }
			 else{
				 termLeft += p * probKGapAmbiguity(t->getLeftChild(), column, i);
			 }
		}
	}
	double termRight =0;
	if(t->getRightChild()->isLeaf()){
		char yChar= column.at(t->getRightChild()->getAlignmentRaw());
		if(yChar=='-'){
			termRight=1;
		}
		else{
			unsigned int y= charToInt(yChar);
			termRight =t->getRightChild()->getPt()->getTable()[x][y];
		}
	}else{
		for(unsigned int i=0;i<4;i++){
			double p=0;
			p=t->getRightChild()->getPt()->getTable()[x][i];
			 if(t->getRightChild()->getProbK().at(i)!=-1){
				 termRight += p * t->getRightChild()->getProbK().at(i);
			 }
			 else{
				 termRight += p * probKGapAmbiguity(t->getRightChild(), column, i);
			 }
		}
	}
	t->getProbK().at(x)=(termLeft * termRight);
	return (termLeft * termRight);
}

double Likelihood::probColumnPair(Node * const & root, string const & columnA,string const & columnB, Matrix * const & m_pairedProbs ){
	resetprobK(root);
	double stateProb;
	double wColumn=0;
	unsigned int firstBase=0;
	//reihenfolge: aa ac ag au ca cc cg cu ga gc gg gu ua uc ug uu
	for(unsigned int i=0;i<4;i++){
		for(unsigned int j=0;j<4;j++){
			stateProb=m_pairedProbs->getTable()[i][j];
			wColumn +=  stateProb* probKPairGapAmbiguity(root, columnA,columnB, firstBase+j);
		}
		firstBase+=4;
	}
	double gapCounter2=0;
		for(unsigned int j=0;j<columnA.size();j++){
			if((columnA.at(j)=='-' && columnB.at(j)!='-') || (columnA.at(j)!='-' && columnB.at(j)=='-')){
				gapCounter2++;
			}
		}
		double epsi = epsilon*wColumn *(gapCounter2/columnA.size());
		return (wColumn-epsi);
}

double Likelihood::probColumn(Node * const & root, string const & column, vector<double> * const & m_unpairedProbs){
	resetprobK(root);
	double stateProb;
	double wColumn=0;
	//reihenfolge: a, c, g ,u
	for(unsigned int i=0; i<4;i++){
		stateProb=m_unpairedProbs->at(i);
		wColumn +=  stateProb* probKGapAmbiguity(root, column, i);		
	}
	return wColumn;

}

void Likelihood::calculateLikelihood(map<string, double> * const & columnLikelihood, map<string, double> * const & columnPairLikelihood,
		Node * const & node, vector<string> * const alignment, vector<double> * const & m_unpairedProbs,
		Matrix * const & m_pairedProbs){
	for(unsigned int i=0; i<alignment->size()-1;i++){
		(*columnLikelihood) [alignment->at(i)] = probColumn(node,alignment->at(i),m_unpairedProbs);
		for(unsigned int j=i+1; j<alignment->size();j++){
			string a=alignment->at(i)+alignment->at(j);
			(*columnPairLikelihood) [a] = probColumnPair(node, alignment->at(i), alignment->at(j), m_pairedProbs);
			}
	}
	(*columnLikelihood) [alignment->at(alignment->size()-1)] = probColumn(node,alignment->at(alignment->size()-1),m_unpairedProbs);
}

void Likelihood::printLikelihoods(map<string, double> * const & columnLikelihood, map<string, double> * const & columnPairLikelihood){
	//cout<<" columnProb:   Elemente: "<<columnLikelihood->size()<<endl;
	map<string, double>::iterator iter;
	double sum=0;
	double sumPair=0;
	for(iter = columnLikelihood->begin(); iter != columnLikelihood->end();iter++){
		sum += iter->second;
	}
	//cout<<"sum: "<<sum<<endl;
	//cout<<" columnPairProb:  Elemente: "<<columnPairLikelihood->size() <<endl;
	for(iter = columnPairLikelihood->begin(); iter != columnPairLikelihood->end();iter++){
		//cout<<iter->first<<" : "<<iter->second<<endl;
		sumPair += iter->second;
	}
	//cout<<"sum: "<<sum<<endl;
	//cout<<"sumPair: "<<sumPair<<endl;
}

