
/****************************************************************************
 * This file is part of the program  PhyloQFold 			                *
 * for MCMC-sampling consensus secondary structures with pseudoknot         *
 * for a given alignment of RNA sequences and a given phylogeny.            *                                       
 *                                                                          *
 *    Copyright (C) 2012  Dirk Metzler & Gero Doose                        *
 *    e-mail: metzler@bio.lmu.de and gero.doose@gmx.de                      *
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/


#ifndef LIKELIHOOD_H_
#define LIKELIHOOD_H_
#include <map>
#include "Matrix.h"
#include "Node.h"
#include <iostream>
#include <fstream>
#include <cstdlib>


class Likelihood {
private:
	//konstanter parameter der angiebt wie stark gaps innerhalb von stem regionen bestraft werden sollen (von 0 fuer garnicht bis 1 fuer sehr stark)
	double epsilon;
	int charToInt(char const a);
	void readEpsilon(const char * const dateiname);
	void resetprobK(Node * const &  r);
	char intToChar(int const a);
	int charToIntPair(string const a);
	string intToCharPair(int const a);
	double probKGapAmbiguity(Node * const & t, string const & column, unsigned int const x );
	double probKPairGapAmbiguity(Node * const & t, string  const & columnA, string const & columnB, unsigned int x );
	double probColumnPair(Node * const & root, string const & columnA,string const & columnB, Matrix * const & m_pairedProbs );
	double probColumn(Node * const & root, string const & column, vector<double> * const & m_unpairedProbs);
public:
	Likelihood( map<string, double> * const & columnLikelihood,map<string,
			double> * const & columnPairLikelihood,Node * const & n,vector<string> * const alignment,
			vector<double> * const & m_unpairedProbs,
			Matrix * const & m_pairedProbs, const char * const dateiname);
	virtual ~Likelihood();
	void calculateLikelihood(map<string, double> * const & columnLikelihood, map<string, double> * const & columnPairLikelihood,
				Node * const & node, vector<string> * const alignment, vector<double> * const & m_unpairedProbs,
				Matrix * const & m_pairedProbs);
	void printLikelihoods(map<string, double> * const & columnLikelihood, map<string, double> * const & columnPairLikelihood);
};


#endif /* LIKELIHOOD_H_ */
