
/****************************************************************************
 * This file is part of the program  PhyloQFold 			                *
 * for MCMC-sampling consensus secondary structures with pseudoknot         *
 * for a given alignment of RNA sequences and a given phylogeny.            *                                       
 *                                                                          *
 *    Copyright (C) 2012  Dirk Metzler & Gero Doose                        *
 *    e-mail: metzler@bio.lmu.de and gero.doose@gmx.de                      *
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#include "Matrix.h"


Matrix::Matrix(unsigned int n):
m(n)
{
	//allocate speicher table
	table = new (double* [m]);
	for (unsigned int i = 0; i < m; i++) {
		table[i] = new double[m];
	}
}

 Matrix::~Matrix(){
	//speicher wieder frei geben
	unsigned int i;
	for (i = 0; i < m; i++) {
		delete table[i];
	}
	delete table;
}

double ** & Matrix::getTable(){
	return table;
}

 unsigned int Matrix::getM()const{
	return m;
}

 void Matrix::ausgeben()const {
 	unsigned int i, j;
 	for (i = 0; i < m; i++) {
 		for (j = 0; j < m; j++) {
 			cout << table[i][j] << "\t";
 		}
 		cout << endl;
 	}
 }

 void Matrix::diagonalEintraegeAngleichen(){
	double diagonal=0;
	unsigned int i, j;
	for (i = 0; i < m; i++) {
		diagonal=0;
		for (j = 0; j < m; j++) {
			if(i!=j){
				diagonal+=table[i][j];
			}
		}
		for (j = 0; j < m; j++) {
			if(i==j){
				table[i][j] =-diagonal;
			}
		}
	}
  }

 void Matrix::ausgebenZeilenSumme() {
 	unsigned int i, j;
 	for (i = 0; i < m; i++) {
 		double z=0;
 		for (j = 0; j < m; j++) {
 			z+=table[i][j];
 		}
 		cout<<"zeilensumme "<<i<<" = "<< z<<endl;
 		cout << endl;
 	}
 }
