
/****************************************************************************
 * This file is part of the program  PhyloQFold 			                *
 * for MCMC-sampling consensus secondary structures with pseudoknot         *
 * for a given alignment of RNA sequences and a given phylogeny.            *                                       
 *                                                                          *
 *    Copyright (C) 2012  Dirk Metzler & Gero Doose                        *
 *    e-mail: metzler@bio.lmu.de and gero.doose@gmx.de                      *
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/


#include "Node.h"

int Node::elems=0;

Node::~Node() {
	elems--;
}
Node::Node( double dist,bool b):
	distance(dist), leaf(b)
{
	pt =new Matrix(4);
	ptPaired = new Matrix (16);
	elems++;
}
vector<double> & Node::getProbK(){
	return probK;
}
vector<double> & Node::getProbKpair(){
	return probKpair;
}
void Node::setAlignmentRaw(int r){
	alignmentRaw=r;
}
int Node::getAlignmentRaw(){
	return alignmentRaw;
}
void Node::setName(string s){
	name=s;
}
string Node::getName(){
	return name;
}
 double Node::getDistance(){
	return distance;
}
Node * const & Node::getLeftChild(){
	return leftChild;
}
Node * const & Node::getRightChild(){
	return rightChild;
}
void Node::setLeftChild(Node * const & left){
	leftChild=left;
}
void Node::setRightChild(Node * const & right){
	rightChild = right;
}
bool Node::isLeaf(){
	return leaf;
}
void Node::setLeaf(bool b){
	leaf=b;
}
Matrix * const & Node::getPt(){
	return pt;
}
Matrix * const & Node::getPtPaired(){
	return ptPaired;
}
int Node::amount(){
	return elems;
}
