
/****************************************************************************
 * This file is part of the program  PhyloQFold 			                *
 * for MCMC-sampling consensus secondary structures with pseudoknot         *
 * for a given alignment of RNA sequences and a given phylogeny.            *                                       
 *                                                                          *
 *    Copyright (C) 2012  Dirk Metzler & Gero Doose                        *
 *    e-mail: metzler@bio.lmu.de and gero.doose@gmx.de                      *
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/



#ifndef NODE_H_
#define NODE_H_
#include <string>
#include <vector>
#include <cstdlib>
#include "Matrix.h"
using std::string;
class Node {

private:

	//wahrscheinlichkeit der daten (ueber diesem knoten)
	vector<double> probK;
	vector<double> probKpair;
	//abbildung aufs alignment
	int alignmentRaw;
	string  name;
	//distanz/zeit zum naechsten Elter
	 double distance;
	 //rekursionsanker
	 bool leaf;
	Node * leftChild;
	Node * rightChild;
	//uebergans wahrscheinlichkeiten der nukleotid(e)||(paare) gegeben der distanz zum naechsten elter.
	Matrix *  pt;
	Matrix * ptPaired;
	//nur zum testen einen instanzen zaehler
	static int elems;

public:
	virtual ~Node();
	Node( double distance, bool b);
	vector<double> & getProbK();
	vector<double> & getProbKpair();
	int getAlignmentRaw();
	void setAlignmentRaw(int);
	void setLeaf(bool b);
	bool isLeaf();
	string getName();
	void setName( string s);
	double getDistance();
	//Tree * getLeftChild();
	Node * const & getLeftChild();
	Node * const & getRightChild();
	void setLeftChild(Node * const & left);
	void setRightChild(Node * const & right);
	Matrix * const & getPt();
	Matrix * const & getPtPaired();
	 static int amount();
};

#endif /* NODE_H_ */
