
/****************************************************************************
 * This file is part of the program  PhyloQFold 			                *
 * for MCMC-sampling consensus secondary structures with pseudoknot         *
 * for a given alignment of RNA sequences and a given phylogeny.            *                                       
 *                                                                          *
 *    Copyright (C) 2012  Dirk Metzler & Gero Doose                        *
 *    e-mail: metzler@bio.lmu.de and gero.doose@gmx.de                      *
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#include <fstream>
#include <map>
#include <set>
#include <string>
#include <cmath>
#include <cstdlib>

#include "samplesec.hh"
#include "mhmcmc.hh"
#include "Node.h"
#include "Matrix.h"
#include "EvoMod.h"
#include "Likelihood.h"
#include "Phylogeny.h"


int main(int argc, char** argv) {
    if (argc==1) {
	unsigned t=time(NULL);
	srand(t);
    }
    else {
	unsigned int seed;
	sscanf(argv[1],"%u",&seed);
	srand(seed);
    }
    vector <string> seqIds;
    vector <string> * const pAlignment= new vector <string>;
    //laden des alignments
    Phylogeny phy(pAlignment,seqIds);
    //laden des baumes,  generierung der tree datenstruktur
    //verknuepfen der fasta header und der leaf labels
    Node * const root = phy.parseNewick(phy.loadtree(),seqIds);
    vector<double> * m_unpairedProbs=new vector<double>;
	Matrix * const m_pairedProbs = new Matrix(4);
	Matrix * const m_unpairedRate = new Matrix(4);
	Matrix * const m_pairedRate = new Matrix(16);
	//ausgehend der raten matritzen  und des baumes
	//berechnung der uebergangs matritzen einer jeden kante.
	EvoMod evoMod(root, m_unpairedProbs, m_pairedProbs, m_unpairedRate, m_pairedRate);
	delete(m_unpairedRate);
	delete(m_pairedRate);
	map <string, double>  * const columnLikelihood = new map <string, double>();
	map <string, double>  * const columnPairLikelihood = new map <string, double>();
	// berechnung der likelihood des baumes bezogen auf die einzelnen spalten und spalten paare
	Likelihood like(columnLikelihood, columnPairLikelihood,root,pAlignment, m_unpairedProbs, m_pairedProbs,"input_grammar");
	delete(m_unpairedProbs);
	delete(m_pairedProbs);
	delete(root);
	Psknmh mcmcsampler(columnLikelihood, columnPairLikelihood, pAlignment, "input_grammar");
    vector<string> ausgabe=mcmcsampler.sample(20,50,100);
    ofstream ausgfile("output_samples");
    for (size_t i=0; i<ausgabe.size(); ++i) {
		cout << ausgabe[i] << endl;
		ausgfile << ausgabe[i] << endl;
    }
    mcmcsampler.saveposteriors("output_pairprobs");
    ausgfile.close();
    string ergebnis=mcmcsampler.simul_anneal(1000);
    cout<<endl;
    for(unsigned int i=0;i<pAlignment->at(0).size();i++){
    	  for(unsigned int j=0;j<pAlignment->size();j++){
    		  cout<<pAlignment->at(j).at(i);
    	  }
    	  cout<<endl;
    }
    cout<<endl<< ergebnis << endl;
    ofstream ergfile("output_folding");
    ergfile << ergebnis << endl;
    ergfile.close();
    delete(pAlignment);
    return 0;
}
