
/****************************************************************************
 * This file is part of the program  PhyloQFold 			                *
 * for MCMC-sampling consensus secondary structures with pseudoknot         *
 * for a given alignment of RNA sequences and a given phylogeny.            *                                       
 *                                                                          *
 *    Copyright (C) 2012  Dirk Metzler & Gero Doose                        *
 *    e-mail: metzler@bio.lmu.de and gero.doose@gmx.de                      *
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/



#include "Phylogeny.h"


Phylogeny::Phylogeny(vector<string>* const & pAlignment, vector<string> & seqIds) {
	loadAlignment(pAlignment,seqIds);
}
//uberladener constructor
//fuer simulationsstudie  n ist die anzahlder sequencen in dem zu simulierenden alignment
//Phylogeny::Phylogeny(vector<string>* const & alignmentForColProb,
//		 vector<string> & seqIds,unsigned int n){
//	loadAlignment(alignmentForColProb,seqIds,n);
//}

Phylogeny::~Phylogeny() {
}

char Phylogeny::intToChar(int const a){
	char i;
	switch(a){
	case 0:
		i='a';
		break;
	case 1:
		i='c';
		break;
	case 2:
		i='g';
		break;
	case 3:
		i='u';
		break;

	}
	return i;
}
void Phylogeny::loadAlignment(vector <string> * const  & pAlignment,vector <string> & seqIds ){
	ifstream file;
	string buf;
	file.open("input_alignment");
	if (!file.is_open()) {
		throw "Error loading File (input_alignment)";
	}
	int z=0;
	while (std::getline(file, buf)){
		if (!buf.empty()){
		   if(buf.at(0)!='>'){
			   for(unsigned int i=0;i<buf.size();i++){
				 if(z==0){
					(*pAlignment).push_back(string());
				 }
				(*pAlignment).at(i).push_back(buf[i]);
			   }
			 z++;
		   }
		   else{
			   seqIds.push_back(buf.substr(1));
		   }
		}
	}
	file.close();
}
Node * const Phylogeny::buildTree (string & newick, Node * const & parent, int from, int to, vector <string> const & seqIds) {
	  if (newick.at(from) != '(') {
		  if(newick.substr(from,to-from).at(newick.substr(from,to-from).length()-1)!=')'){
			  parent->setName(newick.substr(from,to-from));
		  }
		  else{
			  parent->setName(newick.substr(from,(to-from)-1));
		  }
		  parent->setLeaf(true);
		  //zuordnung der blätter zu den zugehörigen alignment zeilen
		  for(unsigned int s=0; s<seqIds.size();s++){
			  if(parent->getName()==seqIds.at(s)){
				  parent->setAlignmentRaw(s);
			  }
		  }
		  for(unsigned int h=0; h<4;h++){
			  parent->getProbK().push_back(-1);
		  }
		  for(unsigned int g=0; g<16;g++){
			  parent->getProbKpair().push_back(-1);
		  }
		  return parent;
	  }
	  int brackets = 0;
	  int beginDistanceLeft=0;
	  for (int i = from; i < to; i++) {
		  char c = newick.at(i);
		  if (c == '(')
			  brackets++;
		  else if (c == ')')
			  brackets--;
		  else if (c == ':')
			  beginDistanceLeft = i+1;
		  if (brackets == 0 || brackets == 1 && c == ',') {
			  int beginDistanceRight= (newick.substr(0, to)).find_last_of(':')+1;
			  parent->setLeftChild(buildTree(newick,(new Node(atof((newick.substr(beginDistanceLeft,
					  i-(beginDistanceLeft))).data()),false)),from+1,beginDistanceLeft-1, seqIds));
			 //die abfrage noch ausmerzen
			  if((newick.at(beginDistanceRight)-2)!=')'){
				  parent->setRightChild(buildTree(newick,(new Node(atof((newick.substr(beginDistanceRight,
						  to-(beginDistanceRight))).data()),false)),i+1,beginDistanceRight-1, seqIds));
			  }
			  else{
				  parent->setRightChild(buildTree(newick,(new Node(atof((newick.substr(beginDistanceRight,
						  (to-(beginDistanceRight))-1)).data()),false)),i+1,beginDistanceRight-1,seqIds));

			  }
			  i = to;
		  }
	  }
	  for(unsigned int h=0; h<4;h++){
		  parent->getProbK().push_back(-1);
	  }
	  for(unsigned int g=0; g<16;g++){
		  parent->getProbKpair().push_back(-1);
	  }
	  return parent;
  }

string Phylogeny::loadtree(){
		ifstream file;
		file.open("input_tree");
		if (!file.is_open()) {
			throw "Error loading File (input_tree)";
		}
		string newick;
		file >> newick;
		file.close();
		cout<<newick<<endl<<endl ;
		return newick;
}
Node * const Phylogeny::parseNewick(string s,vector <string> const & seqIds) {
	if(s.at(s.size()-1)==';')
		s.erase(s.size()-1);
	int a=s.find_last_of(':');
	int b=s.find_last_of(')');
	if(a>b){
	  Node * pRoot = new Node(atof(s.substr(a+1).data()),false);
	  return buildTree(s, pRoot, 0, a, seqIds);
	}
	else{
	  Node * pRoot = new Node(0.0,false);
	  return buildTree(s, pRoot, 0, b, seqIds);
	}
}
