
/****************************************************************************
 * This file is part of the program  PhyloQFold 			                *
 * for MCMC-sampling consensus secondary structures with pseudoknot         *
 * for a given alignment of RNA sequences and a given phylogeny.            *                                       
 *                                                                          *
 *    Copyright (C) 2012  Dirk Metzler & Gero Doose                        *
 *    e-mail: metzler@bio.lmu.de and gero.doose@gmx.de                      *
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#include "mhmcmc.hh"



double Qlogwkeit(set<Hsp> vorschlag,double q) {
    double ergebnis=0.0;
    for(set<Hsp>::iterator it=vorschlag.begin();it!=vorschlag.end(); ++it)
    	ergebnis+=it->logemisswkeit+log(q)*(it->len-3)+log(1-q);
    for (int i=2*vorschlag.size()-1; i>0; i-=2)
    	ergebnis-=log((double)i);
    return ergebnis;
}

bool operator<(const Hsp & a, const Hsp & b) {
    if (a.x<b.x) 
	return true;
    if (a.x>b.x)
	return false;
    if (a.y<b.y) 
	return true;
    if (a.y>b.y)
	return false;
    if (a.len < b.len)
	return true;
    return false;
}


bool Psknmh::neuer_vorschlag() {
    double r=rand()/(RAND_MAX+1.0);
    vorschlag=aktuell;
    vorschlagswkeitenquotient=1.0;
    qwkeitquot=1.0;
    kombfakt=1.0;
    //cout << "current number of q-stems: " << vorschlag.size() << endl;
    if (r<0.5) {
		// Loesche ein hsp falls moeglich
		if (vorschlag.size()>0) {
			r=rand()/(RAND_MAX+1.0)*(vorschlag.size())-1;
			set<Hsp>::iterator it=vorschlag.begin();
			for (it=vorschlag.begin(); r>0 && it!=vorschlag.end(); ++it) {
				r-=1;
			}
			if (it==vorschlag.end()) {
				cerr << "r= " << r << endl;
				cerr << "iteration out of range in void psknmh::neuer_vorschlag()\n";
				exit(1);
			}
			//cout << "Proposal: remove q-Stem: " << (*it).x << "-"
			 //<< (*it).x+(*it).len-1 << " with "
			 //<< (*it).y<<  "-"  << (*it).y+(*it).len-1 << endl;
			//cout << "wscore: " << (*it).wscore << " ,    score: " << (*it).score
			 //<< " ,   wkeit : " << exp(it->logemisswkeit)  << endl;
			vorschlagswkeitenquotient=vorschlag.size()*(*it).wscore/secsamp.get_totalwsc();
			qwkeitquot=1/(exp(it->logemisswkeit)*pow(secsamp.getqparam(),((*it).len-3))*(1-secsamp.getqparam()));
			kombfakt=2*vorschlag.size()-1;
			vorschlag.erase(it);
			return true;
		}
    }
    else {
		// Waehle ein hsp dazu
		Hsp diag=secsamp.randhsp();
		//cout << "proposed q-Stem: " << diag.x << "-"  << diag.x+diag.len-1
			// << " with " << diag.y<<  "-"  << diag.y+diag.len-1 << endl;
		size_t m=vorschlag.size();
		vorschlag.insert(diag);
		if (vorschlag.size()>m) {
			vorschlagswkeitenquotient=1/((m+1)*(diag.wscore/secsamp.get_totalwsc()));
			//cout << vorschlagswkeitenquotient << " " << m << " " << diag.wscore << " " << secsamp.get_totalwsc() << endl;
			qwkeitquot=exp(diag.logemisswkeit)*pow(secsamp.getqparam(),(diag.len-3))*(1-secsamp.getqparam());
			kombfakt=1.0/(2*vorschlag.size()-1);
			// cout << "Vorschlagsgroesse " << vorschlag.size() << endl;
			return true;
		}
    }
    return false;
}

//--------------------------------------------------------------

void Psknmh::akzeptanz_schritt() {
    vector<size_t> qreg(0);
    //cout << vorschlag.size() << endl;
    for(set<Hsp>::iterator it=vorschlag.begin();it!=vorschlag.end(); ++it) {
		qreg.push_back(it->x);
		qreg.push_back(it->x+it->len-1);
		qreg.push_back(it->y);
		qreg.push_back(it->y+it->len-1);
    }
    Sampler structsamp_vorschlag(columnLikelihood, columnPairLikelihood, originalAlignment, wkeitendatei, qreg);
    double r=rand()/(RAND_MAX+1.0);
    //     cout << "Vorschlag fuer die Q-Regionen:\n";
    //     for (size_t i=0;i<qreg.size()/2;i+=2)
    // 	cout << qreg[i] << '\t' << qreg[i+1] << endl;
    if( structsamp_vorschlag.logwkeit()+Qlogwkeit(vorschlag, structsamp_vorschlag.getqparam()) > maxlogwk ) {
		argmaxhsps=vorschlag;
		maxlogwk=structsamp_vorschlag.logwkeit()+Qlogwkeit(vorschlag,structsamp_vorschlag.getqparam());
    }
//     cout << "vorschlagswkeitenquotient " << vorschlagswkeitenquotient << endl;
//     cout << "logvorschl " <<  structsamp_vorschlag.logwkeit() << "      logaktuell " << structsamp_aktuell.logwkeit() << endl;
//     cout << "qwkeitquot " << qwkeitquot << endl;
//     cout << "kombinatorischer Faktor: " << kombfakt << endl;
    double akzeptwkeit=vorschlagswkeitenquotient
	* exp(structsamp_vorschlag.logwkeit()-structsamp_aktuell.logwkeit())
	* qwkeitquot * kombfakt;
//     cout << "Akzeptanzwahrscheinlichkeit = " << (1<akzeptwkeit?1:akzeptwkeit) << endl;
    if(protokollieren) {
	protfile << "mcmc\t" << structsamp_aktuell.logwkeit()+Qlogwkeit(aktuell,structsamp_aktuell.getqparam()) << '\t' << aktuell.size() << '\t' 
		 << structsamp_vorschlag.logwkeit()+Qlogwkeit(vorschlag,structsamp_vorschlag.getqparam()) << '\t' << vorschlag.size() << '\t'
		 << maxlogwk <<  '\t' << argmaxhsps.size()
		 << '\t' << exp(structsamp_vorschlag.logwkeit()-structsamp_aktuell.logwkeit()) * qwkeitquot * kombfakt
		 << '\t' <<  akzeptwkeit
		 << endl;
	//cout << "mcmc\t" << structsamp_aktuell.logwkeit()+Qlogwkeit(aktuell,structsamp_aktuell.getqparam()) << '\t' << aktuell.size() << '\t'
	  //   << structsamp_vorschlag.logwkeit()+Qlogwkeit(vorschlag,structsamp_vorschlag.getqparam()) << '\t' << vorschlag.size() << '\t'
	    // << maxlogwk <<  '\t' << argmaxhsps.size()
	     //<< '\t' << exp(structsamp_vorschlag.logwkeit()-structsamp_aktuell.logwkeit()) * qwkeitquot * kombfakt
	     //<< '\t' <<  akzeptwkeit
	     //<< endl;
    }
    if( r< akzeptwkeit ){
    	structsamp_aktuell=structsamp_vorschlag;
		aktuell=vorschlag;
		//cout << "Proposal accepted\n\n";
    	//cout<<"r = "<<r<<" ist kleiner als  akzeptwkeit= "<<akzeptwkeit<<endl;

    }
    else{
    	//cout << "Proposal rejected\n\n";
    	//cout<<"r = "<<r<<" ist groesser als  akzeptwkeit= "<<akzeptwkeit<<endl;
    }
}

//---------------------------------------------------------------
//wie im sampleteil nur ohne vorschlagswahrscheinlichkeitsquotient und mit temperaturfaktor
void Psknmh::sa_akzeptanz_schritt(double einsdurchtemperatur) {
    vector<size_t> qreg(0);
    //cout << vorschlag.size() << endl;
    for(set<Hsp>::iterator it=vorschlag.begin();it!=vorschlag.end(); ++it) {
		qreg.push_back(it->x);
		qreg.push_back(it->x+it->len-1);
		qreg.push_back(it->y);
		qreg.push_back(it->y+it->len-1);
    }
    Sampler structsamp_vorschlag(columnLikelihood, columnPairLikelihood,
    		 originalAlignment, wkeitendatei, qreg);
    double r=rand()/(RAND_MAX+1.0);
    //     cout << "Vorschlag fuer die Q-Regionen:\n";
    //     for (size_t i=0;i<qreg.size()/2;i+=2)
    // 	cout << qreg[i] << '\t' << qreg[i+1] << endl;
    if( structsamp_vorschlag.logwkeit()+Qlogwkeit(vorschlag, structsamp_vorschlag.getqparam()) > maxlogwk ) {
		argmaxhsps=vorschlag;
		maxlogwk=structsamp_vorschlag.logwkeit()+Qlogwkeit(vorschlag,structsamp_vorschlag.getqparam());
    }
    //cout << "Number q-Stems: " << vorschlag.size() << endl;
    //cout << "Total prob. of all foldings with these q-stems" << endl;
     //cout << "logvorschl " <<  structsamp_vorschlag.logwkeit() << "      logaktuell " << structsamp_aktuell.logwkeit() << endl;
     //cout << "qwkeitquot " << qwkeitquot << endl;
     //cout << "kombinatorischer Faktor: " << kombfakt << endl;

    double akzeptwkeit=exp(einsdurchtemperatur*(structsamp_vorschlag.logwkeit()-structsamp_aktuell.logwkeit()))
	* pow(qwkeitquot * kombfakt,einsdurchtemperatur);
    //cout << "acceptance probability = " << akzeptwkeit << endl;

    if(protokollieren) {
	protfile << "sa\t" << structsamp_aktuell.logwkeit()+Qlogwkeit(aktuell,structsamp_aktuell.getqparam()) << '\t' << aktuell.size() << '\t' 
		 << structsamp_vorschlag.logwkeit()+Qlogwkeit(vorschlag,structsamp_vorschlag.getqparam()) << '\t' << vorschlag.size() << '\t'
		 << maxlogwk <<  '\t' << argmaxhsps.size() 
		 << '\t' << exp((structsamp_vorschlag.logwkeit()-structsamp_aktuell.logwkeit()))*qwkeitquot * kombfakt
		 << '\t' << akzeptwkeit
		 << endl;
	//cout << "sa\t" << structsamp_aktuell.logwkeit()+Qlogwkeit(aktuell,structsamp_aktuell.getqparam()) << '\t' << aktuell.size() << '\t'
	  //   << structsamp_vorschlag.logwkeit()+Qlogwkeit(vorschlag,structsamp_vorschlag.getqparam()) << '\t' << vorschlag.size() << '\t'
	    // << maxlogwk <<  '\t' << argmaxhsps.size()
	     //<< '\t' << exp((structsamp_vorschlag.logwkeit()-structsamp_aktuell.logwkeit()))*qwkeitquot * kombfakt
	     //<< '\t' << akzeptwkeit
	     //<< endl;
    }
    if( r< akzeptwkeit ){
	structsamp_aktuell=structsamp_vorschlag;
	aktuell=vorschlag;
	//cout << "Proposal accepted\n\n";
    }
    //else cout << "Proposal rejected\n\n";
}

//---------------------------------------------------------------

void Psknmh::sa_max_akzeptanz_schritt(double einsdurchtemperatur) {
    // Hier wird nur die wahrscheinlichste Faltung fuer die gegebene
    // Menge an Q-Stems betrachtet. 
    vector<size_t> qreg(0);
    //cout << vorschlag.size() << endl;
    for(set<Hsp>::iterator it=vorschlag.begin();it!=vorschlag.end(); ++it) {
	qreg.push_back(it->x);
	qreg.push_back(it->x+it->len-1);
	qreg.push_back(it->y);
	qreg.push_back(it->y+it->len-1);
    }
   // cout << "Proposal:\n";
    Cyk vorschlag_cyk(aktuell_cyk,qreg);
    double r=rand()/(RAND_MAX+1.0);
    //     cout << "Vorschlag fuer die Q-Regionen:\n";
    //     for (size_t i=0;i<qreg.size()/2;i+=2)
    // 	cout << qreg[i] << '\t' << qreg[i+1] << endl;
    if( vorschlag_cyk.get_logwkeit()+Qlogwkeit(vorschlag, vorschlag_cyk.get_qparam()) > maxmaxlogwk ) {
	argmaxmaxhsps=vorschlag;
	maxmaxlogwk=vorschlag_cyk.get_logwkeit()+Qlogwkeit(vorschlag,vorschlag_cyk.get_qparam());
    }
    //cout << "Number of q-stems: " << vorschlag.size() << endl;
     //cout << "Wahrscheinlichkeiten der wahrscheinlichsten Faltungen mit diesen Q-Stems" << endl;
     //cout << "logvorschl " <<  vorschlag_cyk.get_logwkeit() << "      logaktuell " << aktuell_cyk.get_logwkeit() << endl;
     //cout << "qwkeitquot " << qwkeitquot << endl;
    //cout << "kombinatorischer Faktor: " << kombfakt << endl;
    double akzeptwkeit=exp(einsdurchtemperatur*(vorschlag_cyk.get_logwkeit()-aktuell_cyk.get_logwkeit()))
	* pow(qwkeitquot * kombfakt,einsdurchtemperatur);
   // cout << "acceptance probability = " << akzeptwkeit<< endl;
    if(protokollieren) {
	protfile << "sa_max\t" << aktuell_cyk.get_logwkeit()+Qlogwkeit(aktuell,aktuell_cyk.get_qparam()) << '\t' << aktuell.size() << '\t' 
		 << vorschlag_cyk.get_logwkeit()+Qlogwkeit(vorschlag,vorschlag_cyk.get_qparam()) << '\t' << vorschlag.size() << '\t'
		 << maxmaxlogwk <<  '\t' << argmaxmaxhsps.size() 
		 << '\t' << exp((vorschlag_cyk.get_logwkeit()-aktuell_cyk.get_logwkeit())) * qwkeitquot * kombfakt
		 <<  '\t' << akzeptwkeit << endl;
	//cout << "sa_max\t" << aktuell_cyk.get_logwkeit()+Qlogwkeit(aktuell,aktuell_cyk.get_qparam()) << '\t' << aktuell.size() << '\t'
	  //   << vorschlag_cyk.get_logwkeit()+Qlogwkeit(vorschlag,vorschlag_cyk.get_qparam()) << '\t' << vorschlag.size() << '\t'
	    // << maxmaxlogwk <<  '\t' << argmaxmaxhsps.size()
	    // << '\t' << exp(vorschlag_cyk.get_logwkeit()-aktuell_cyk.get_logwkeit()) * qwkeitquot * kombfakt <<  '\t' << akzeptwkeit << endl;
    }
    if( r< akzeptwkeit ){
	aktuell_cyk=vorschlag_cyk;
	aktuell=vorschlag;
	//cout << "proposal accepted\n\n";
    }
    //else cout << "proposal rejected\n\n";
}

//---------------------------------------------------------------

string Psknmh::zuordne(string struktur, set<Hsp> hsps) {
    // ersetzt die Nullen durch Klein- und Grossbuchstaben, so dass 
    // sich die Pseudoknotenstemhaelften einander zuordnen lassen
    const string klammer("aAbBcCdDeEfFgGhHiIjJkKlLmMnNoOpPqQrRsStTuUvVwWxXyYzZ");
    size_t j=0;
    for (set<Hsp>::iterator it=hsps.begin(); j+1<klammer.size() && it!=hsps.end(); ++it) {
		for(int k=it->x; k<it->x+it->len; ++k)
			struktur[k]=klammer[j];
		for(int k=it->y; k<it->y+it->len; ++k)
			struktur[k]=klammer[j+1];
		j+=2;
    }
    return struktur;
}

//--------------------------------------------------------------


Psknmh::Psknmh(map <string, double>  * const & columnLikelihood,map <string, double>  * const & columnPairLikelihood,
		 vector<string > * const & rpAlignment, const char * const datei):
	columnLikelihood(columnLikelihood),columnPairLikelihood(columnPairLikelihood),
    originalAlignment(rpAlignment),
    aktuell_cyk(columnLikelihood,columnPairLikelihood,originalAlignment),
    secsamp(columnLikelihood, columnPairLikelihood, originalAlignment, datei),
    structsamp_aktuell(secsamp),
    paarwkeiten(originalAlignment->size(),vector<double>(originalAlignment->size(),0.0)),
    wkeitendatei(datei),
    maxlogwk(-numeric_limits<double>::max()),
    maxmaxlogwk(-numeric_limits<double>::max())
{
};

//--------------------------------------------------------------




vector<string> Psknmh::sample(int anzahl, int steps, int burnin) {
    vector<string> ausgabe;
    if(protokollieren)
	protfile.open("logfile");
    for (size_t n=0; n<originalAlignment->size(); ++n){
		for (size_t m=0; m<n; ++m){
			paarwkeiten[n][m]=0.0;
		}
    }
    for (int i=0;i<burnin; ++i) {
		cout << "burn-in step " << i+1 << " of " << burnin << endl;
		if (neuer_vorschlag())
			akzeptanz_schritt();
    }

    for (int i=0;i<anzahl; ++i) {
		for (int j=0;j<steps; ++j) {
			cout << "step " << j+1 << " of " << steps << " for sample " << i+1 << " of " << anzahl << endl;
			if (neuer_vorschlag())
				akzeptanz_schritt();
		}
		//randomisiertes Backtracking einer moeglichen Faltung
		string s=structsamp_aktuell.sample();
		ausgabe.push_back(zuordne(s,aktuell));
		//cout << zuordne(s,aktuell) << endl << endl;
		for (size_t n=0; n<originalAlignment->size(); ++n)
			for (size_t m=0; m<originalAlignment->size(); ++m)
				//innerhalb der methode get_posteriwpaar wird einmal der outside algo aufgerufen,
				//wenn fuer das aktuelle sampler objekt (der 20 stueck)der pasteriwpaar vector <vector <double>> noch leer sein sollte.
				paarwkeiten[n][m]+=structsamp_aktuell.get_posteriwpaar(n,m)/anzahl;
		for (set<Hsp>::iterator it=aktuell.begin(); it!=aktuell.end(); ++it){
			for (int z=0; z<it->len; ++z) {
				paarwkeiten[it->x+z][it->y+it->len-1-z]+=1.0/anzahl;
				paarwkeiten[it->y+it->len-1-z][it->x+z]+=1.0/anzahl;
			}
		}
		//ausgabe.push_back(zuordne(s,aktuell));
    }  
    if(protokollieren)
    	protfile.close();
    cout << "\nMCMC-samplig finished\n";
    return ausgabe;
}

//-----------------------------------------------------------------------------------
    
string  Psknmh::simul_anneal (int steps) {
    if(protokollieren)
	protfile.open("logfile",ios_base::app);
   //beginne mit der im sample teil am wahrscheinlichsten bewerteten q stem konfiguration (hoechste posterior probability)
    aktuell=argmaxhsps;
    vector<size_t> qreg(0);
    cout << "\nFirst Simulated-Annealing phase\n";
    //in qreg die anfang und end positionen der q stems eintragen
    for(set<Hsp>::iterator it=aktuell.begin();it!=aktuell.end(); ++it) {
		qreg.push_back((*it).x);
		qreg.push_back(it->x+it->len-1);
		qreg.push_back(it->y);
		qreg.push_back(it->y+it->len-1);
    }
    structsamp_aktuell.set_qreg(qreg);
    for (int i=0; i<steps; ++i) {
		cout << i << endl;
		if(i%50==0)
			cout << structsamp_aktuell.sample() << endl;
		if (neuer_vorschlag())
			sa_akzeptanz_schritt(i/300.0);
    }
    aktuell=argmaxhsps;
    qreg.resize(0);
    cout << "\nSecond Simulated-Annealing phase\n";
    for(set<Hsp>::iterator it=aktuell.begin();it!=aktuell.end(); ++it) {
		qreg.push_back(it->x);
		qreg.push_back(it->x+it->len-1);
		qreg.push_back(it->y);
		qreg.push_back(it->y+it->len-1);

    }
    //auf dem Cyk objekt die methode setzte(Sampler, Vector) anwenden
	//intern wird falte() aufgerufen
    aktuell_cyk.setz(structsamp_aktuell,qreg);
    argmaxmaxhsps=aktuell;
    maxmaxlogwk=aktuell_cyk.get_logwkeit()+Qlogwkeit(aktuell,aktuell_cyk.get_qparam());
    for (int i=0; i<steps; ++i) {
		cout << i << endl;
		if(i%50==0)
			cout << zuordne(aktuell_cyk.get_faltung(),aktuell) << endl;
		if (neuer_vorschlag())
			sa_max_akzeptanz_schritt(i/300.0);
    }
    secsamp.loesche_maps();
    structsamp_aktuell.loesche_maps();
    qreg.resize(0);
    for(set<Hsp>::iterator it=argmaxmaxhsps.begin();it!=argmaxmaxhsps.end(); ++it) {
		qreg.push_back(it->x);
		qreg.push_back(it->x+it->len-1);
		qreg.push_back(it->y);
		qreg.push_back(it->y+it->len-1);
    }
    Cyk ergebnis_cyk(aktuell_cyk,qreg);
    if(protokollieren)
	protfile.close();
    cout << "\nSimulated Annealing finished\n\n";
    cout << "Found folding of log(W'keit) " 
	 << ergebnis_cyk.get_logwkeit()+Qlogwkeit(argmaxmaxhsps,ergebnis_cyk.get_qparam()) << endl;
    return zuordne(ergebnis_cyk.get_faltung(),argmaxmaxhsps);
}
