
/****************************************************************************
 * This file is part of the program  PhyloQFold 			                *
 * for MCMC-sampling consensus secondary structures with pseudoknot         *
 * for a given alignment of RNA sequences and a given phylogeny.            *                                       
 *                                                                          *
 *    Copyright (C) 2012  Dirk Metzler & Gero Doose                        *
 *    e-mail: metzler@bio.lmu.de and gero.doose@gmx.de                      *
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/
#include <iostream>
#include <fstream>
#include <vector>
#include <set>
#include <map>
#include <iterator>
#include <ctime>
#include <string>
#include <cmath>
#include <cstdlib>
#include <limits>
#include "samplesec.hh"
class Psknmh {
private :
	map <string, double>  * const columnLikelihood;
	map <string, double>  * const columnPairLikelihood;
	vector <string > * originalAlignment;

	Cyk aktuell_cyk;
	Sampler secsamp, structsamp_aktuell;
    // secsamp ist der sampler fuer sekundaerstrukturen ohne pseudoknots

    set<Hsp> aktuell, vorschlag, argmaxhsps, argmaxmaxhsps;
    vector<vector<double> > paarwkeiten;
    const char * const wkeitendatei;
    double vorschlagswkeitenquotient, // Qy->x/Qx->y
	qwkeitquot,
	kombfakt,
	maxlogwk,
	maxmaxlogwk;
    bool neuer_vorschlag();
    void akzeptanz_schritt();
    void sa_akzeptanz_schritt(double einsdurchtemperatur);
    void sa_max_akzeptanz_schritt(double einsdurchtemperatur);
    string zuordne(string struktur,set<Hsp> hsps);
    // ersetzt die Nullen durch Klein- und Grossbuchstaben, so dass 
    // sich die Pseudoknotenstemhaelften einander zuordnen lassen
    const static bool protokollieren=true;
    ofstream protfile;
public :
    Psknmh(map <string, double>  * const & columnLikelihood,map <string, double>  * const & columnPairLikelihood,
    		 vector<string > * const & rpAlignment, const char * const dateiname);
    vector<string> sample(int anzahl, int steps, int burnin); 
    string simul_anneal(int steps); 

    void saveposteriors(const char * datei) {
		ofstream dat(datei);
		for (size_t i=0;i<paarwkeiten.size();++i)
			for (size_t j=0;j<paarwkeiten[i].size();++j)
			dat << paarwkeiten[i][j] << endl;
		dat.close();
    }
};
