
/****************************************************************************
 * This file is part of the program  PhyloQFold 			                *
 * for MCMC-sampling consensus secondary structures with pseudoknot         *
 * for a given alignment of RNA sequences and a given phylogeny.            *                                       
 *                                                                          *
 *    Copyright (C) 2012  Dirk Metzler & Gero Doose                        *
 *    e-mail: metzler@bio.lmu.de and gero.doose@gmx.de                      *
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/



#include "samplesec.hh"

/*  Grammatik:
 *
 *  S -> LS | x | axFyb | Q
 *  F -> xLSy | xFy  
 *  L -> x | axFyb | Q
 */



class Faltungseintrag {
	private:
		string anfang;
		pair<int,int> A,B;
		string ende;
		char cA,cB;
	public:
		Faltungseintrag() :
		anfang(""), A(-1,-1), B(-1,-1), ende("") {
		cA='X';
		cB='X';
		}
		Faltungseintrag(string s) :
		anfang(s), A(-1,-1), B(-1,-1), ende("") {
		cA='X';
		cB='X';
		}
		Faltungseintrag(char c,int x, int y) :
		anfang(""), A(x,y), B(-1,-1), ende("") {
		cA=c;
		cB='X';
		}
		Faltungseintrag(string r, char c,int x, int y,  string s) :
		anfang(r), A(x,y), B(-1,-1), ende(s) {
		cA=c;
		cB='X';
		}
		Faltungseintrag(string r, char c,int a, int b,  char z, int x, int y, string s) :
		anfang(r), A(a,b), B(x,y), ende(s) {
		cA=c;
		cB=z;
		}
		Faltungseintrag( char c,int a, int b,  char z, int x, int y) :
		anfang(""), A(a,b), B(x,y), ende("") {
			cA=c;
			cB=z;
		}
		inline void setze(string s) {
			anfang=s;
			A.first=-1;
			A.second=-1;
			cA='X';
			cB='X';
			B.first=-1;
			B.second=-1;
			ende=-1;
		}
		inline void setze(string r, char c, int x, int y, string s) {
			anfang=r;
			A.first=x;
			A.second=y;
			cA=c;
			cB='X';
			B.first=-1;
			B.second=-1;
			ende=s;
		}
		inline void setze(char c,int a, int b){
			anfang="";
			A.first=a;
			A.second=b;
			cA=c;
			cB='X';
			B.first=-1;
			B.second=-1;
			ende="";
		}
		inline void setze(char c,int a, int b,char z, int x, int y){
			anfang="";
			A.first=a;
			A.second=b;
			cA=c;
			cB=z;
			B.first=x;
			B.second=y;
			ende="";
		}
		inline void setze(string r, char c,int a, int b,char z, int x, int y,  string s){
			anfang=r;
			A.first=a;
			A.second=b;
			cA=c;
			cB=z;
			B.first=x;
			B.second=y;
			ende=s;
		}
		inline string get_anfang() const {return anfang;}
		inline string get_ende() const {return ende;}
		inline int get_A1() const {return A.first;}
		inline int get_A2() const {return A.second;}
		inline char get_cA() const {return cA;}
		inline int get_B1() const {return B.first;}
		inline int get_B2() const {return B.second;}
		inline char get_cB() const {return cB;}
};

class Rekursive_faltung {
private:
    int sequenzlaenge;
    map<char,vector<vector<Faltungseintrag> > > beste_faltung;
    string ausgabe(char c, int i, int j) {
		string A(""),B("");
		Faltungseintrag f= beste_faltung[c].at(i).at(j);
		if (f.get_A1()>=0)
			A=ausgabe(f.get_cA(),f.get_A1(),f.get_A2());
		if (f.get_B1()>=0)
			B=ausgabe(f.get_cB(),f.get_B1(),f.get_B2());
		return f.get_anfang()+A+B+f.get_ende();
    }
public:
    Rekursive_faltung(int n) {
		sequenzlaenge=n;
		vector<vector<Faltungseintrag> > w(0);
		beste_faltung['S']=w;
		beste_faltung['F']=w;
		beste_faltung['L']=w;
		beste_faltung['Q']=w;
		beste_faltung['S'].resize(n);
		beste_faltung['F'].resize(n);
		beste_faltung['L'].resize(n);
		beste_faltung['Q'].resize(n);
		for (int i=0; i<n; ++i) {
			beste_faltung['S'][i].resize(n-i);
			beste_faltung['F'][i].resize(n-i);
			beste_faltung['L'][i].resize(n-i);
			beste_faltung['Q'][i].resize(n-i);
		}
    }
    inline void setze(char c, int i, int j, Faltungseintrag f) {
    	beste_faltung[c][i][j]=f;
    }
    inline string ausgabe() {
    	return ausgabe('S',0,sequenzlaenge-1);
    }
};


/*****************************************
 *        class sampler                  *
 *****************************************/

Sampler::Sampler( map <string, double>  * const & columnLikelihood,map <string, double>  * const & columnPairLikelihood,
	 vector<string > * const & rpAlignment,
	 const char * wkeitendatei,const vector<size_t> & qreg) :
		columnLikelihood(columnLikelihood),columnPairLikelihood(columnPairLikelihood),
		originalAlignment(rpAlignment),
		mAlignment(*originalAlignment),
		qregion(0),diag(),alpha(),beta(),posteriwpaar(0)
{
    // Der Vektor qregion enthaelt abwechselnd Anfang und Ende von Q-Regionen
    lieswkeiten(wkeitendatei);
    set_qreg(qreg);
}

Sampler::Sampler( map <string, double>  * const & columnLikelihood,map <string, double>  * const & columnPairLikelihood,
		vector<string > * const & rpAlignment, const char * const wkeitendatei) :
		columnLikelihood(columnLikelihood),columnPairLikelihood(columnPairLikelihood),
		originalAlignment(rpAlignment),mAlignment(*rpAlignment),
		qregion(0),diag(),alpha(),beta(),posteriwpaar(0)
{
    const double minusinf=-numeric_limits<double>::max();
    lieswkeiten(wkeitendatei);

    size_t n=mAlignment.size();
    vector<vector<double> > v;
    alpha['S']=v;
    alpha['F']=v;
    alpha['L']=v;
    alpha['Q']=v;
    alpha['S'].resize(n);
    alpha['F'].resize(n);
    alpha['L'].resize(n);
    alpha['Q'].resize(n);
    for (unsigned int i=0; i<n; ++i) {
		alpha['S'][i].resize(n-i);
		alpha['F'][i].resize(n-i);
		alpha['L'][i].resize(n-i);
		alpha['Q'][i].resize(n-i);
		for (unsigned int j=0; j<n-i; ++j) {
			alpha['S'][i][j]=minusinf;
			alpha['F'][i][j]=minusinf;
			alpha['L'][i][j]=minusinf;
			alpha['Q'][i][j]=minusinf;
		}
    }
    inside_algo();
    /*
    cout<<" alpha Werte:   Elemente: "<<alpha.size()<<endl;
	map<char, vector<vector<double> > >::iterator iter;
	for(iter = alpha.begin(); iter != alpha.end();iter++){
		cout<<iter->first<<endl;
		for(unsigned int i=0; i<iter->second.size();i++){
			for(unsigned int j=0; j<iter->second.at(i).size();j++){
				if(iter->second.at(i).at(j)!=minusinf)
				cout<<" : "<<iter->second.at(i).at(j);
			}
		}
		cout<<endl;
	}
	*/
}

void Sampler::outside_algo() {
    // berechnet beta mit outside-Algo
    const double minusinf=-numeric_limits<double>::max();
    int n=(int)mAlignment.size();
    vector<vector<double> > v;
    vector<vector<double> > logwpaar;
    beta['S']=v;
    beta['F']=v;
    beta['L']=v;
    beta['Q']=v;
    beta['S'].resize(n);
    beta['F'].resize(n);
    beta['L'].resize(n);
    beta['Q'].resize(n);
    logwpaar.resize(n);
    for (int i=0; i<n; ++i) {
		beta['S'][i].resize(n-i);
		beta['F'][i].resize(n-i);
		beta['L'][i].resize(n-i);
		beta['Q'][i].resize(n-i);
		logwpaar[i].resize(n);
		for (int j=0; j<n-i; ++j) {
			beta['S'][i][j]=minusinf;
			beta['F'][i][j]=minusinf;
			beta['L'][i][j]=minusinf;
			beta['Q'][i][j]=minusinf;
		}
		for (int j=0; j<n; ++j)
			logwpaar[i][j]=minusinf;
    }
    beta['S'][0][n-1]=0.0;
    for(int dist=n-1; dist>=0; --dist) {
		for(int i=0;i+dist<n; ++i) {
			int j=i+dist;
			// Regel S -> Q     (dist >= 0)
			beta['Q'][i][dist]=addlog(beta['Q'][i][dist],beta['S'][i][dist]+logStoQ);

			// Regel L -> Q     (dist >= 0)
			beta['Q'][i][dist]=addlog(beta['Q'][i][dist],beta['L'][i][dist]+logLtoQ);

			// Regeln S -> x und L -> x sind hier irrelevant
			//if (i+1<=dist)
			if (dist>0) {
				// Regel S -> LS  (dist >= 1)
				double w=beta['S'][i][dist]+logStoLS;
				for(int k=i; k<j; ++k) {
					beta['L'][i][k-i]=addlog(beta['L'][i][k-i],w+alpha['S'][k+1][j-k-1]);
					beta['S'][k+1][j-k-1]=addlog(beta['S'][k+1][j-k-1],w+alpha['L'][i][k-i]);
				}
				if (dist>1) {
					double logwp;
					if(mAlignment.at(i).at(mAlignment.at(i).size()-1)=='0'|| mAlignment.at(j).at(mAlignment.at(j).size()-1)=='0'){
						logwp=minusinf;
					}
					else{
						string s= mAlignment.at(i)+mAlignment.at(j);
						logwp=log((*columnPairLikelihood)[s]);
					}
					// Regel F -> xFy   (dist >= 2)
					beta['F'][i+1][dist-2]=addlog(beta['F'][i+1][dist-2],beta['F'][i][dist]+logFtoxFy+logwp);
					// siehe ww
					logwpaar[i][j]=addlog(logwpaar[i][j],beta['F'][i][dist]+logFtoxFy+logwp+alpha['F'][i+1][dist-2]);
					if (dist>2) {
						// Regel F -> xLSy  (dist >= 3)
						w=beta['F'][i][dist]+logFtoxLSy+logwp;
						double ww=minusinf;
						for(int k=i+1; k<j-1; ++k) {
							beta['L'][i+1][k-i-1]=addlog(beta['L'][i+1][k-i-1],w+alpha['S'][k+1][j-k-2]);
							beta['S'][k+1][j-k-2]=addlog(beta['S'][k+1][j-k-2],w+alpha['L'][i+1][k-i-1]);
							//noch mal ueberpruefen: ww als argument von addlog, ww kann doch auch minus_inf sein. 
							ww=addlog(ww,alpha['L'][i+1][k-i-1]+alpha['S'][k+1][j-k-2]);
						}
						logwpaar[i][j]=addlog(logwpaar[i][j],ww+w);
						if (dist>3) {
							double logwp2;
							if(mAlignment.at(i+1).at(mAlignment.at(i+1).size()-1)=='0'|| mAlignment.at(j-1).at(mAlignment.at(j-1).size()-1)=='0'){
								logwp2=minusinf;
							}
							else{
								string s2=mAlignment.at(i+1)+mAlignment.at(j-1);
								logwp2=log((*columnPairLikelihood)[s2]);
							}
							// Regel L -> axFyb (dist >= 4)
							beta['F'][i+2][dist-4]=addlog(beta['F'][i+2][dist-4],beta['L'][i][dist]+logLtoaxFyb+logwp+logwp2);
							logwpaar[i][j]=addlog(logwpaar[i][j],beta['L'][i][dist]+logLtoaxFyb+logwp+logwp2+alpha['F'][i+2][dist-4]);
							logwpaar[i+1][j-1]=addlog(logwpaar[i+1][j-1],beta['L'][i][dist]+logLtoaxFyb+logwp+logwp2+alpha['F'][i+2][dist-4]);

							// Regel S -> axFyb (dist >= 4)
							beta['F'][i+2][dist-4]=addlog(beta['F'][i+2][dist-4],beta['S'][i][dist]+logStoaxFyb+logwp+logwp2);
							logwpaar[i][j]=addlog(logwpaar[i][j],beta['S'][i][dist]+logStoaxFyb+logwp+logwp2+alpha['F'][i+2][dist-4]);
							logwpaar[i+1][j-1]=addlog(logwpaar[i+1][j-1],beta['S'][i][dist]+logStoaxFyb+logwp+logwp2+alpha['F'][i+2][dist-4]);
						}
					}
				}
			}
		}
    }
    posteriwpaar.resize(n);
    for(int i=0;i<n;++i) {
    	posteriwpaar[i].resize(n);
    }
    for(int i=0;i<n;++i) {
		posteriwpaar[i][i]=0;
		for(int j=i+1; j<n; ++j) {
			posteriwpaar[i][j]=exp(logwpaar[i][j]-alpha['S'][0][n-1]);
			posteriwpaar[j][i]=posteriwpaar[i][j];
			//cout << posteriwpaar[i][j] << ' ' << logwpaar[i][j] << endl ;
		}
    }
}


void Sampler::inside_algo() {
    int i,j,k,dist,n=mAlignment.size();
    double x,wp0,wp1,wp2;
    for (i=0; i<n; ++i) {
    	if((*mAlignment.at(i).end())!='0'){
    		wp0=(*columnLikelihood)[mAlignment.at(i)];
    		// Regel S -> x
			alpha['S'][i][0]=log(wp0)+logStox;
			// Regel L -> x
			alpha['L'][i][0]=log(wp0)+logLtox;
    	}
    	//wenn an position i der sequence eine Q region startet und nur EINS lang ist:
		if (alpha['Q'][i][0]>-1) {
			// Regel S -> Q falls Q aktiv
			alpha['S'][i][0]=logStoQ;
			// Regel L -> Q falls Q aktiv
			alpha['L'][i][0]=logLtoQ;
		}
    }
    for (dist=1; dist<n; ++dist) {
		for (i=0; i+dist<n; ++i) {
			j=i+dist;
			if (alpha['Q'][i][dist]>-1) {
				// Regel S -> Q falls Q aktiv
				alpha['S'][i][dist]=logStoQ;
				// Regel L -> Q falls Q aktiv
				alpha['L'][i][dist]=logLtoQ;
			}
			// Regel S -> LS
			for (k=i;k<j;++k) {
				alpha['S'][i][j-i]=addlog(alpha['S'][i][j-i],logStoLS+alpha['L'][i][k-i]+alpha['S'][k+1][j-k-1]);
			}
			// Regel F -> xLSy
			if((*mAlignment.at(i).end())=='0' || (*mAlignment.at(j).end())=='0'){
				wp1= 0;
			}
			else{
				string s=mAlignment.at(i)+mAlignment.at(j);
				wp1=(*columnPairLikelihood)[s];
			}
			if (wp1>0){
				for (k=i+1;k<j-1;++k){
					alpha['F'][i][j-i]=addlog(alpha['F'][i][j-i],
							logFtoxLSy+log(wp1)+alpha['L'][i+1][k-i-1]+alpha['S'][k+1][j-k-2]);
				}
			}
			if (i+2<=j) {
				// Regel F -> xFy
				if (wp1>0){
					alpha['F'][i][j-i]=addlog(alpha['F'][i][j-i],logFtoxFy+log(wp1)+alpha['F'][i+1][j-i-2]);
				}
				if (i+4<=j) {
					if((* mAlignment.at(i+1).end())=='0' || (*mAlignment.at(j-1).end())=='0'){
						wp2=0;
					}
					else{
						string s2=mAlignment.at(i+1)+mAlignment.at(j-1);
						wp2=(*columnPairLikelihood)[s2];
					}
					if (wp1>0 && wp2>0) {
						x=log(wp1)+log(wp2);
						// Regel S -> axFyb
						alpha['S'][i][j-i]=addlog(alpha['S'][i][j-i],logStoaxFyb+x+alpha['F'][i+2][j-i-4]);
						// Regel L -> axFyb
						alpha['L'][i][j-i]=addlog(alpha['L'][i][j-i],logLtoaxFyb+x+alpha['F'][i+2][j-i-4]);
					}
				}
			}
		}
    }
}

void stimmts(string s, string s2) {
	//cerr << "s: "<<s<<" s2: "<<s2<<endl;
    	if (s!=s2) {
	    cerr << "Error in file model_probs" << endl;
	    cerr << "s: "<<s<<" s2: "<<s2<<endl;
	    exit(1);
	}
}

void Sampler::alignmentNew() {
    // Findet und scort alle HSPs (ab einer Lange von 3 bp)
    // der Sequenz mit ihrem Komplement
    // und weist ihnen einen gewichteten Score zu.
    const int n= mAlignment.size();
    vector<vector<double> > s(n);
    vector<vector<int> > leng(n), maxsl(n);
    //testen des zweiten criteriums benoetigt die posteriwahrscheinlichkeiten fuer alle paarungen der basen der sequenz
    if (posteriwpaar.size()==0)
    	outside_algo();
    for (int i=0; i<n; ++i) {
		s[i].resize(n,1.0);
		leng[i].resize(n,0);
		maxsl[i].resize(n,0);
    }
    for (int j=0; j<n; ++j) {
    	string help=mAlignment.at(0)+mAlignment.at(j);
    	double p=(*columnPairLikelihood)[help]/
    			(*columnLikelihood)[mAlignment.at(0)]/(*columnLikelihood)[mAlignment.at(j)];
    	//p=log(p);
		if (p>=1) {
			s[0][j]=p;
			leng[0][j]=1;
			maxsl[0][j]=1;
		}
		help=mAlignment.at(j)+mAlignment.at(n-1);
		p=(*columnPairLikelihood)[help]/
				(*columnLikelihood)[mAlignment.at(n-1)]/(*columnLikelihood)[mAlignment.at(j)];
		//p=log(p);
		if (p>=1) {
			s[j][n-1]=p;
			leng[j][n-1]=1;
			maxsl[j][n-1]=1;
		}
    }
    for (int i=1; i<n; ++i) {
		for (int j=i+1; j<n-1; ++j) {
			string help=mAlignment.at(i)+mAlignment.at(j);
			double p=(*columnPairLikelihood)[help]/
					(*columnLikelihood)[mAlignment.at(i)]/(*columnLikelihood)[mAlignment.at(j)];
			//p=log(p);
			if (p>1.000001) {
				s[i][j]=p*s[i-1][j+1];
				leng[i][j]=1+leng[i-1][j+1];
				int d=leng[i][j]-maxsl[i-1][j+1];
				if (s[i][j]>s[i-d][j+d])
					maxsl[i][j]=leng[i][j];
				else
					maxsl[i][j]=maxsl[i-1][j+1];
			}
			if (p<1.000001  && leng[i-1][j+1]>=3) {
				int x=i-leng[i-1][j+1],
					y=j+leng[i-1][j+1],
					m=maxsl[i-1][j+1];
				double sc=s[x+m-1][y-m+1];
						//cout<<"s i j = "<<sc<<" leng "<<leng.at(i-1).at(j+1)<<endl;
				double q=0.0001;
				double lemwk=0.0;
				for (int k=0;k<m;++k) {
					string help=mAlignment.at(x+k)+mAlignment.at(y-k);
					double p=(*columnPairLikelihood)[help];
					lemwk += log(p);
					//cout<<"k= "<<k<<" lemwk: "<<lemwk<<"  q= "<<q<<" posteriwpaar "<<x+k<<" "<<y-k<<" = "<<posteriwpaar.at(x+k).at(y-k)<<endl;
					if (q<posteriwpaar[x+k][y-k])
						q=posteriwpaar[x+k][y-k];
				}
				Hsp loula(x,y-m+1,m,sc,(1-exp(-0.000001*sc))/q,lemwk);
				diag.push_back(loula);
			}
		}
		if (leng[i-1][i]>=3) {
			int x=i-leng[i-1][i],
			y=i-1+leng[i-1][i],
			m=maxsl[i-1][i];
			double sc=s[x+m-1][y-m+1],
			q=0.0001;
			double lemwk=0.0;
			for (int k=0;k<m;++k) {
				string help=mAlignment.at(x+k)+mAlignment.at(y-k);
				double p=(*columnPairLikelihood)[help];
				lemwk += log(p);
				if (q<posteriwpaar[x+k][y-k])
					q=posteriwpaar[x+k][y-k];
			}
			Hsp loula(x,y-m+1,m,sc,(1-exp(-0.000001*sc))/q,lemwk);
			diag.push_back(loula);
		}
    }
    totalwsc=0.0;
    for (size_t i=0; i<diag.size(); ++i) {
    	totalwsc+=diag[i].wscore;
    }
    cout << "------------------------------------------------------------------------------------\n";
    cout << "Possible q-stems, their score, their truncated SCFG-Probabiliry and their weight\n";
    for (size_t i=0; i<diag.size(); ++i) {
		cout << diag[i].x << "-" << diag[i].x+diag[i].len-1 << "\t mit "
		<< diag[i].y << "-" << diag[i].y+diag[i].len-1 << '\t'
		<< diag[i].score << '\t'
		<< (1-exp(-0.000001*diag[i].score))/diag[i].wscore << '\t'
		<< diag[i].wscore/totalwsc << endl;
    }
    cout << "------------------------------------------------------------------------------------\n";
}
void Sampler::lieswkeiten(const char * const dateiname){
    
    ifstream f(dateiname);
	string s;
	f >> s;
	stimmts(s,"Epsilon");
	f >> s;

	f >> s;
	stimmts(s,"StoLS");
	f >> StoLS;
	f >> s;
	stimmts(s,"Stox");
	f >> Stox;
	f >> s;
	stimmts(s,"StoaxFyb");
	f >> StoaxFyb;
	f >> s;
	stimmts(s,"StoQ");
	f >> StoQ;
	if(abs(StoLS+Stox+StoaxFyb+StoQ-1.0)>1e-10) {
	    cerr << "Error in File " << dateiname << " : Sum=1+" << StoLS+Stox+StoaxFyb+StoQ-1<< endl;
	    exit(1);
	}
	f >> s;
	stimmts(s,"FtoxLSy");
	f >> FtoxLSy;
	f >> s;
	stimmts(s,"FtoxFy");
	f >> FtoxFy;
	if(abs(FtoxLSy+FtoxFy-1.0)>1e-10) {
	    cerr << "Error in File " << dateiname << " : Sum=1+" <<FtoxLSy+FtoxFy-1 <<endl;
	    exit(1);
	}
	f >> s;
	stimmts(s,"Ltox");
	f >> Ltox;
	f >> s;
	stimmts(s,"LtoaxFyb");
	f >> LtoaxFyb;
	f >> s;
	stimmts(s,"LtoQ");
	f >> LtoQ;
	if(abs(Ltox+LtoaxFyb+LtoQ-1.0)>1e-10) {
	    cerr << "Error in File " << dateiname << " : Sum=1+" <<  Ltox+LtoaxFyb+LtoQ-1<<endl;
	    exit(1);
	}
	f >> s;
	stimmts(s,"q");
	f >> qparam;
	logStoLS=log(StoLS);
	logStox=log(Stox);
	logStoaxFyb=log(StoaxFyb);
	logStoQ=log(StoQ);
	logFtoxFy=log(FtoxFy);
	logFtoxLSy=log(FtoxLSy);
	logLtox=log(Ltox);
	logLtoaxFyb=log(LtoaxFyb);
	logLtoQ=log(LtoQ);
	logqparam=log(qparam);
};


void mecker(int n, int i, int j, char X) {

    cerr << "Error " << n << " in string sampler::sample(int i, int j, char X)" << i << " " << j << " " << X << endl;
    exit(1);
}
//randomisiertes Backtracking einer moeglichen Faltung
//beschleunigte version
string Sampler::sample(int i, int j, char X) {
    // sample ausgehend von non-Terminal X eine Faltung fuer 
    // sequenz[i],...,sequenz[j], wobei Q als Q ausgegeben wird.
    string ausgabe="";
    if (j<i)
    	mecker(1,i,j,X);
    if (X=='Q') {
		for (int k=i; k<=j; ++k){
			ausgabe.push_back('0');
		}
    }
    else {
		if (j==i) {
			if(X=='S' || X=='L') {
				if ((*mAlignment.at(i).end())=='0'){
					ausgabe.push_back('0');
				}
				else{
					// Regeln S -> x und L-> x
					ausgabe.push_back('.');
				}
			}
			else mecker(2,i,j,X);
		}
		if (j==i+1) {
			if(X=='S') {
				// Regel S -> Q
				//Q, startposition, laenge
				if(alpha['Q'][i][j-i]>-1) {
					ausgabe+=sample(i,j,'Q');
				}
				else {
					// Regel S -> LS
					ausgabe+=sample(i,i,'L');
					ausgabe+=sample(j,j,'S');
				}
			}
			else
				mecker(3,i,j,X);
		}
		if (j>i+1) {
			if (X=='L') {
				// Regel L -> Q
				if (alpha['Q'][i][j-i]>-1) {
					ausgabe+=sample(i,j,'Q');
				}
				else {
					// Regel L -> axFyb

					//diese ganzen abfragen sind eigentlich auch nicht mehr noetig, oder?
					double wp1,wp2;

						string s=mAlignment.at(i)+mAlignment.at(j);
						wp1=(*columnPairLikelihood)[s];

						string s2=mAlignment.at(i+1)+mAlignment.at(j-1);
						wp2=(*columnPairLikelihood)[s2];

					if (wp1>0 && wp2>0 )
						ausgabe+=(string)"(("+=sample(i+2,j-2,'F')+=(string)"))";
					else
						mecker(4,i,j,X);
				}
			}
			if (X=='F') {
				double wp1;

					string s=mAlignment.at(i)+mAlignment.at(j);
					wp1=(*columnPairLikelihood)[s];

				if(wp1<=0)
					mecker(5,i,j,X);
				double r=rand()/(RAND_MAX+1.0);
				double logWS=alpha['F'][i][j-i];
				bool okay=false;

				// Regel F -> xFy
				r-=exp(logFtoxFy+log(wp1)+alpha['F'][i+1][j-i-2]-logWS);
				if (r<0) {
					ausgabe+=(string)"("+=sample(i+1,j-1,'F')+=(string)")";
					okay=true;
				}

				// Regel F -> xLSy
				for (int k=i+1;k<j-1 && !(r<0);++k) {
					r-=exp(logFtoxLSy+log(wp1)+alpha['L'][i+1][k-i-1]+alpha['S'][k+1][j-k-2]-logWS);
					if (r<0) {
						ausgabe+=(string)"("+=sample(i+1,k,'L')+=sample(k+1,j-1,'S')+=(string)")";
						okay=true;
					}
				}
				if (!okay){
					cout<<"wp1= "<<wp1<<endl;
					mecker(6,i,j,X);

				}
			}
			if (X=='S') {
				// Regel S -> Q
				if(alpha['Q'][i][j-i]>-1) {
					ausgabe+=sample(i,j,'Q');
				}
				else {
					double r=rand()/(RAND_MAX+1.0);
					double logWS=alpha['S'][i][j-i];
					bool okay=false;
					if (j>i+3) {
						// Regel S -> axFyb
						double wp1,wp2;

							string s=mAlignment.at(i)+mAlignment.at(j);
							wp1=(*columnPairLikelihood)[s];

							string s2=mAlignment.at(i+1)+mAlignment.at(j-1);
							wp2=(*columnPairLikelihood)[s2];

						r-=exp(logStoaxFyb+log(wp1)+log(wp2)+alpha['F'][i+2][j-i-4]-logWS);
						if (r<0) {
							ausgabe+=(string)"(("+=sample(i+2,j-2,'F')+=(string)"))";
							okay=true;
						}
					}
					// Regel S -> LS
					for (int k=i;k<j && !(r<0);++k) {
						r-=exp(logStoLS+alpha['L'][i][k-i]+alpha['S'][k+1][j-k-1]-logWS);
						// cout << r << '\t' << k << '\t' << i  << '\t' << j << endl;
						if (r<0) {
							ausgabe+=sample(i,k,'L')+=sample(k+1,j,'S');
							okay=true;
						}
					}
					if (!okay)
						mecker(7,i,j,X);
				}
			}
		}
    }
    return ausgabe;
}

//ersetzt in der Sequenz die buchstaben der positionen an denen q stems liegen durch nullen
//setzt alpha-Qs auf 0.0 (aktiv)
//und ruft  auf diesem sampler objekt den inside algo auf.
void Sampler::set_qreg(const vector<size_t> & qreg) {
    const double minusinf=-numeric_limits<double>::max();
    qregion=qreg;
    bool overlap=false;
    mAlignment=*originalAlignment;
    size_t n=mAlignment.size();
    vector<vector<double> > v;
    alpha['S']=v;
    alpha['F']=v;
    alpha['L']=v;
    alpha['Q']=v;
    alpha['S'].resize(n);
    alpha['F'].resize(n);
    alpha['L'].resize(n);
    alpha['Q'].resize(n);
    for (unsigned int i=0; i<n; ++i) {
		alpha['S'][i].resize(n-i);
		alpha['F'][i].resize(n-i);
		alpha['L'][i].resize(n-i);
		alpha['Q'][i].resize(n-i);
		for (unsigned int j=0; j<n-i; ++j) {
			alpha['S'][i][j]=minusinf;
			alpha['F'][i][j]=minusinf;
			alpha['L'][i][j]=minusinf;
			alpha['Q'][i][j]=minusinf;
		}
    }
    for (unsigned int i=0; i+1<qregion.size(); i+=2) {
    	if (qregion[i]>qregion[i+1] || qregion[i+1]>=n) {
		//if (qregion[i]>qregion[i+1] || qregion[i+1]>=sequenz.size()) {
			cerr << "unallowed q-region\n";
			cerr << qregion[i] << ' ' << qregion[i+1] << ' ' << mAlignment.size() << endl;
			//cerr << qregion[i] << ' ' << qregion[i+1] << ' ' << sequenz.size() << endl;
			exit(1);
		}
		//	cout << "Qregion " << qregion[i] << " " << qregion[i+1] << endl;
		//in der sequenz alle berreiche auf 0 setzten die zu einer der qregion gehören
		for (unsigned int j=qregion[i]; j<=qregion[i+1]; ++j) {
			if (mAlignment.at(j).at(mAlignment.at(j).size()-1)=='0'){
				overlap=true;
			}
			mAlignment.at(j).push_back('0');
		}
		// alpha mit: Q, begin der qregion, länge-1 der qregion
		alpha['Q'][qregion[i]][qregion[i+1]-qregion[i]]=0.0;
		if(qregion[i+1]-qregion[i]==0){
			cout<<"wird doch gesetzt..."<<endl;
			exit(1);
		}
		// cout << "Q " << qregion[i] << " " << qregion[i+1] << endl;
    }
    // cout << sequenz << endl;
    if (!overlap)
    	inside_algo();

};



Hsp Sampler::randhsp() {
    // Gibt ein zufaelliges Element von diag aus, gemaess dem wscore.
    if (diag.size()==0){
    	alignmentNew();
    	//cout<<"randhsp";
    	//exit(0);
    }
    if (diag.size()==0)
    	cerr << "No hsps available!\n";
    double r=rand()/(RAND_MAX+1.0)*totalwsc;
    size_t i;
    for (i=0; r>0 && i<diag.size(); ++i) 
    	r-=diag[i].wscore;
    return diag[i-1];
}

/***************************************************
 *                class cyk                        *
 ***************************************************/

Cyk::Cyk(map <string, double>  * const & columnLikelihood, map <string, double>  * const & columnPairLikelihood,
    	vector<string > * const & rpAlignment):
    	columnLikelihood(columnLikelihood),
    	columnPairLikelihood(columnPairLikelihood),
    	originalAlignment(rpAlignment),
		mAlignment(*originalAlignment)
    	{

}

Cyk::Cyk(const Sampler & S):
		columnLikelihood(S.get_columnLikelihood()),
		columnPairLikelihood(S.get_columnPairLikelihood()),
		originalAlignment(S.get_OriginalAlignment()),
		mAlignment(*originalAlignment),
		qregion(S.get_qregion()),
		qparam(S.getqparam()),
		logStoLS(S.get_logStoLS()), logStox(S.get_logStox()), logStoaxFyb(S.get_logStoaxFyb()),
		logStoQ(S.get_logStoQ()), logFtoxFy(S.get_logFtoxFy()), logFtoxLSy(S.get_logFtoxLSy()),
		logLtox(S.get_logLtox()), logLtoaxFyb(S.get_logLtoaxFyb()), logLtoQ(S.get_logLtoQ()),
		logqparam(S.get_logqparam())
	{
	falteNew();
};

Cyk::Cyk(const Cyk & C, const vector<size_t> & qreg) :
	columnLikelihood(C.columnLikelihood),
	columnPairLikelihood(C.columnPairLikelihood),
	originalAlignment(C.originalAlignment),
	mAlignment(*originalAlignment),
	qregion(qreg),
	qparam(C.qparam),
    logStoLS(C.logStoLS), logStox(C.logStox), logStoaxFyb(C.logStoaxFyb),
    logStoQ(C.logStoQ), logFtoxFy(C.logFtoxFy), logFtoxLSy(C.logFtoxLSy),
    logLtox(C.logLtox), logLtoaxFyb(C.logLtoaxFyb), logLtoQ(C.logLtoQ),
    logqparam(C.logqparam)
{
	falteNew();
}

void Cyk::setz(const Sampler & S, const vector<size_t> & qreg){
    qregion=qreg;
    qparam=S.getqparam();
    logStoLS=S.get_logStoLS(); logStox=S.get_logStox(); logStoaxFyb=S.get_logStoaxFyb(); 
    logStoQ=S.get_logStoQ(); logFtoxFy=S.get_logFtoxFy(); logFtoxLSy=S.get_logFtoxLSy(); 
    logLtox=S.get_logLtox(); logLtoaxFyb=S.get_logLtoaxFyb(); logLtoQ=S.get_logLtoQ();
    logqparam=S.get_logqparam();
    falteNew();
}

/*
void Cyk::falte() {
    map<char,vector<vector<double> > > gamma;
    // Jeder Nichtterminalen X der Grammatik ordnet die map
     // gamma einen vector<vector<double> > zu, so dass
     //gamma[X][i][j] den log der Wahrscheinlichkeit der wahrscheinlichsten
     //Enstehungsgeschichte der Terminalenfolge sequenz[i],sequenz[i+1],...,sequenz[j]
     //aus einer Variablen X ist.
     //
    const double minusinf=-numeric_limits<double>::max();
    int n=sequenz.size();
    Rekursive_faltung best_falt(n);
    // Enthaelt fue alle X,i,j die Wahrscheinlichste Faltung
      //fuer sequenz[i],sequenz[i+1],...,sequenz[j] bei Start in X

    vector<vector<double> > v;
    gamma['S']=v;
    gamma['F']=v;
    gamma['L']=v;
    gamma['Q']=v;
    gamma['S'].resize(n);
    gamma['F'].resize(n);
    gamma['L'].resize(n);
    gamma['Q'].resize(n);
    for (int i=0; i<n; ++i) {
		gamma['S'][i].resize(n-i);
		gamma['F'][i].resize(n-i);
		gamma['L'][i].resize(n-i);
		gamma['Q'][i].resize(n-i);
		for (int j=0; j<n-i; ++j) {
			gamma['S'][i][j]=minusinf;
			gamma['F'][i][j]=minusinf;
			gamma['L'][i][j]=minusinf;
			gamma['Q'][i][j]=minusinf;
		}
    }
    for (unsigned int i=0; i+1<qregion.size(); i+=2) {
		if (qregion[i]>qregion[i+1] || qregion[i+1]>=sequenz.size()) {
			cerr << "unallowed q-region\n";
			cerr << qregion[i] << ' ' << qregion[i+1] << ' ' << sequenz.size() << endl;
			exit(1);
		}
		//	cout << "Qregion " << qregion[i] << " " << qregion[i+1] << endl;
		for (unsigned int j=qregion[i]; j<=qregion[i+1]; ++j) {
			// 	    if (sequenz[j]=='0')
			// 		overlap=true;
			sequenz[j]='0';
		}
		gamma['Q'][qregion[i]][qregion[i+1]-qregion[i]]=0.0;
		best_falt.setze('Q',qregion[i],qregion[i+1]-qregion[i],Faltungseintrag("Q"));
		// cout << "Q " << qregion[i] << " " << qregion[i+1] << endl;
    }
    double x,wp1,wp2;
    vector<set<int> > L_moeglich(n);
    // L_moeglich[i] enthaelt die j <= dist, für die der Bereich
    // zwischen i und j durch L erzeugt sein kann.
    for (int i=0; i<n; ++i) {
		L_moeglich[i].insert(i);
		for (int dist=1; i+dist<n; ++dist) {
			if (gamma['Q'][i][dist]>-1)
			L_moeglich[i].insert(i+dist);
		}
    }
    for (int i=0; i<n; ++i) {
		switch(sequenz[i]) {
		case 'a':
		case 'c':
		case 'g':
		case 'u':
		case 'A':
		case 'C':
		case 'G':
		case 'U':
		case 't':
		case 'T':

			// Regel S -> x
			gamma['S'][i][0]=get_logpr(sequenz[i])+logStox;
			best_falt.setze('S',i,0,Faltungseintrag("."));

			// Regel L -> x
			gamma['L'][i][0]=get_logpr(sequenz[i])+logLtox;
			best_falt.setze('L',i,0,Faltungseintrag("."));

			break;
		case 'Q':
			//cout << "simulated annealing     Q  Case Q  Case Q  Case Q  Case Q  Case Q  Case"<< endl;
			//exit(1);
			// Regel S -> Q
			gamma['S'][i][0]=logStoQ;
			best_falt.setze('S',i,0,Faltungseintrag("Q"));

			// Regel L -> Q
			gamma['L'][i][0]=logLtoQ;
			best_falt.setze('L',i,0,Faltungseintrag("Q"));
		}
		if (gamma['Q'][i][0]>-1) {
			best_falt.setze('Q',i,0,Faltungseintrag("Q"));

			// Regel S -> Q falls Q aktiv
			gamma['S'][i][0]=logStoQ+gamma['Q'][i][0];

			// Regel L -> Q falls Q aktiv
			gamma['L'][i][0]=logLtoQ+gamma['Q'][i][0];
		}
    }

    for (int dist=1; dist<n; ++dist) {
		for (int i=0; i+dist<n; ++i) {
			int j=i+dist;

			// Regel S -> LS
			for (set<int>::iterator k=L_moeglich[i].begin();k!=L_moeglich[i].end();++k) {
			// cout << i << ' ' << (*k) << ' ' << j << ' ' << n << endl;
			if (*k<j) {
				if (logStoLS+gamma['L'][i][(*k)-i]+gamma['S'][(*k)+1][j-(*k)-1]>gamma['S'][i][j-i]) {
					gamma['S'][i][j-i]=logStoLS+gamma['L'][i][(*k)-i]+gamma['S'][(*k)+1][j-(*k)-1];
					best_falt.setze('S',i,j-i,Faltungseintrag('L',i,(*k)-i,'S',(*k)+1,j-(*k)-1));
				}
			}	    }

			// Regel F -> xLSy
			for (set<int>::iterator k=L_moeglich[i+1].begin();k!=L_moeglich[i+1].end();++k) {
			// cout << i << ' ' << (*k) << ' ' << j << endl;
				if (*k<j-1) {
					wp1 = paarwkeit(sequenz[i],sequenz[j]);
					if (wp1>0 && logFtoxLSy+log(wp1)+gamma['L'][i+1][(*k)-i-1]+gamma['S'][(*k)+1][j-(*k)-2]>gamma['F'][i][j-i]) {
						gamma['F'][i][j-i]=logFtoxLSy+log(wp1)+gamma['L'][i+1][(*k)-i-1]+gamma['S'][(*k)+1][j-(*k)-2];
						best_falt.setze('F',i,j-i,Faltungseintrag("(",'L',i+1,(*k)-i-1,'S',(*k)+1,j-(*k)-2,")"));
					}
				}
			}

			if (gamma['Q'][i][dist]>-1) {
				string nullen("0");
				for(int j=0; j<dist; ++j)
					nullen.append("0");
				best_falt.setze('Q',i,dist,Faltungseintrag(nullen));
				// Regel S -> Q falls Q aktiv
				gamma['S'][i][dist]=logStoQ+gamma['Q'][i][dist];
				best_falt.setze('S',i,dist,Faltungseintrag('Q',i,dist));
				// Regel L -> Q falls Q aktiv
				gamma['L'][i][dist]=logLtoQ+gamma['Q'][i][dist];
				best_falt.setze('L',i,dist,Faltungseintrag('Q',i,dist));
				L_moeglich[i].insert(i+dist);
			}

			if (i+2<=j) {
				// Regel F -> xFy
				wp1 = paarwkeit(sequenz[i],sequenz[j]);
				if (wp1>0 && logFtoxFy+log(wp1)+gamma['F'][i+1][j-i-2]>gamma['F'][i][j-i]) {
					gamma['F'][i][j-i]=logFtoxFy+log(wp1)+gamma['F'][i+1][j-i-2];
					best_falt.setze('F',i,j-i,Faltungseintrag("(",'F',i+1,j-i-2,")"));
				}

				if (i+4<=j) {
					wp1 = paarwkeit(sequenz[i],sequenz[j]);
					wp2 = paarwkeit(sequenz[i+1],sequenz[j-1]);
					if (wp1>0 && wp2>0) {
						x=log(wp1)+log(wp2);

						// Regel S -> axFyb
						if(logStoaxFyb+x+gamma['F'][i+2][j-i-4]>gamma['S'][i][j-i]) {
							gamma['S'][i][j-i]=logStoaxFyb+x+gamma['F'][i+2][j-i-4];
							best_falt.setze('S',i,j-i,Faltungseintrag("((",'F',i+2,j-i-4,"))"));
						}
						// Regel L -> axFyb
						if(logLtoaxFyb+x+gamma['F'][i+2][j-i-4]>gamma['L'][i][j-i]) {
							gamma['L'][i][j-i]=logLtoaxFyb+x+gamma['F'][i+2][j-i-4];
							best_falt.setze('L',i,j-i,Faltungseintrag("((",'F',i+2,j-i-4,"))"));
						}
					if (gamma['L'][i][j-i]>minusinf)
						L_moeglich[i].insert(j);
					}
				}
			}
		}
    }
//     for (int i=0;  i<n; i++)
// 	for (int j=0; j+i<n; j++)
// 	    cout << i << ' ' << j << ' ' << gamma['S'][i][j] << ' '
// 		 << gamma['F'][i][j] << ' ' << gamma['L'][i][j] << ' '
// 		 << gamma['Q'][i][j] << endl;
    logwkeit=gamma['S'][0][n-1];
    faltung=best_falt.ausgabe();
    //return faltung=best_falt.ausgabe();
}
*/

void Cyk::falteNew() {
    map<char,vector<vector<double> > > gamma;
    /* Jeder Nichtterminalen X der Grammatik ordnet die map
     * gamma einen vector<vector<double> > zu, so dass
     * gamma[X][i][j] den log der Wahrscheinlichkeit der wahrscheinlichsten
     * Enstehungsgeschichte der Terminalenfolge sequenz[i],sequenz[i+1],...,sequenz[j]
     * aus einer Variablen X ist.
     *
     * Zusatz: in wahrheit(fuer die hier verwendeten i und j's) bis [i+j]
     */
    const double minusinf=-numeric_limits<double>::max();
    unsigned int n=mAlignment.size();
    Rekursive_faltung best_falt(n);
    /* Enthaelt fuer alle X,i,j die Wahrscheinlichste Faltung
     * fuer sequenz[i],sequenz[i+1],...,sequenz[j] bei Start in X
     */
    vector<vector<double> > v;
    gamma['S']=v;
    gamma['F']=v;
    gamma['L']=v;
    gamma['Q']=v;
    gamma['S'].resize(n);
    gamma['F'].resize(n);
    gamma['L'].resize(n);
    gamma['Q'].resize(n);
    for (unsigned int i=0; i<n; ++i) {
		gamma['S'][i].resize(n-i);
		gamma['F'][i].resize(n-i);
		gamma['L'][i].resize(n-i);
		gamma['Q'][i].resize(n-i);
		for (unsigned int j=0; j<n-i; ++j) {
			gamma['S'][i][j]=minusinf;
			gamma['F'][i][j]=minusinf;
			gamma['L'][i][j]=minusinf;
			gamma['Q'][i][j]=minusinf;
		}
    }
    for (unsigned int i=0; i+1<qregion.size(); i+=2) {
		if (qregion[i]>qregion[i+1] || qregion[i+1]>=mAlignment.size()) {
			cerr << "unallowed q-region\n";
			cerr << qregion[i] << ' ' << qregion[i+1] << ' ' << mAlignment.size() << endl;
			exit(1);
		}
		for (unsigned int j=qregion[i]; j<=qregion[i+1]; ++j) {
			mAlignment.at(j).push_back('0');
		}
		gamma['Q'][qregion[i]][qregion[i+1]-qregion[i]]=0.0;
		best_falt.setze('Q',qregion[i],qregion[i+1]-qregion[i],Faltungseintrag("Q"));
    }

    double x,wp0,wp1,wp2;
    for (unsigned int i=0; i<n; ++i) {
    	if(mAlignment.at(i).at(mAlignment.at(i).size()-1)!='0'){
			// Regel S -> x
    		wp0=(*columnLikelihood)[mAlignment.at(i)];
			gamma['S'][i][0]=log(wp0)+logStox;
			best_falt.setze('S',i,0,Faltungseintrag("."));

			// Regel L -> x
			gamma['L'][i][0]=log(wp0)+logLtox;
			best_falt.setze('L',i,0,Faltungseintrag("."));
    	}
		if (gamma['Q'][i][0]>-1) {
			best_falt.setze('Q',i,0,Faltungseintrag("Q"));
			// Regel S -> Q falls Q aktiv
			//in der alten version wurde  +gamma['Q'][i][0] gerechnet aber das dürfte ja eigentlich immer nur 0 sein, oder?
			gamma['S'][i][0]=logStoQ;
			// Regel L -> Q falls Q aktiv
			gamma['L'][i][0]=logLtoQ;
		}
    }
    for (unsigned int dist=1; dist<n; ++dist) {
		for (unsigned int i=0; i+dist<n; ++i) {
			unsigned int j=i+dist;
			if (gamma['Q'][i][dist]>-1) {
				string nullen("0");
				for(unsigned int j=0; j<dist; ++j){
					nullen.append("0");
				}
				best_falt.setze('Q',i,dist,Faltungseintrag(nullen));
				// Regel S -> Q falls Q aktiv
				//in der alten version wurde  +gamma['Q'][i][0] gerechnet aber das dürfte ja eigentlich immer nur 0 sein, oder?
				gamma['S'][i][dist]=logStoQ;
				best_falt.setze('S',i,dist,Faltungseintrag('Q',i,dist));
				// Regel L -> Q falls Q aktiv
				gamma['L'][i][dist]=logLtoQ;
				best_falt.setze('L',i,dist,Faltungseintrag('Q',i,dist));
			}

			//Regel S -> LS
			for (unsigned int k=i;k<j;++k) {
				if (logStoLS+gamma['L'][i][k-i]+gamma['S'][k+1][j-k-1]>gamma['S'][i][j-i]) {
					gamma['S'][i][j-i]=logStoLS+gamma['L'][i][k-i]+gamma['S'][k+1][j-k-1];
					best_falt.setze('S',i,j-i,Faltungseintrag('L',i,k-i,'S',k+1,j-k-1));
				}
			}

			// Regel F -> xLSy
			//wp1 = paarwkeit(sequenz[i],sequenz[j]);
			if(mAlignment.at(i).at(mAlignment.at(i).size()-1)=='0' || mAlignment.at(j).at(mAlignment.at(j).size()-1)=='0'){
				wp1= 0;
			}
			else{
				string s=mAlignment.at(i)+mAlignment.at(j);
				wp1=(*columnPairLikelihood)[s];
			}
			for (unsigned int k=i+1;k<j-1;++k) {
				if (k<j-1) {
					if (wp1>0 && logFtoxLSy+log(wp1)+gamma['L'][i+1][k-i-1]+gamma['S'][k+1][j-k-2]>gamma['F'][i][j-i]) {
						gamma['F'][i][j-i]=logFtoxLSy+log(wp1)+gamma['L'][i+1][k-i-1]+gamma['S'][k+1][j-k-2];
						best_falt.setze('F',i,j-i,Faltungseintrag("(",'L',i+1,k-i-1,'S',k+1,j-k-2,")"));
					}
				}
			}
			if (i+2<=j) {
				// Regel F -> xFy
				//wp1 = paarwkeit(sequenz[i],sequenz[j]);
				if (wp1>0 && logFtoxFy+log(wp1)+gamma['F'][i+1][j-i-2]>gamma['F'][i][j-i]) {
					gamma['F'][i][j-i]=logFtoxFy+log(wp1)+gamma['F'][i+1][j-i-2];
					best_falt.setze('F',i,j-i,Faltungseintrag("(",'F',i+1,j-i-2,")"));
				}

				if (i+4<=j) {
					//wp1 = paarwkeit(sequenz[i],sequenz[j]);
					//wp2 = paarwkeit(sequenz[i+1],sequenz[j-1]);
					if( mAlignment.at(i+1).at(mAlignment.at(i+1).size()-1)=='0' || mAlignment.at(j-1).at(mAlignment.at(j-1).size()-1)=='0'){
						wp2=0;
					}
					else{
						string s2=mAlignment.at(i+1)+mAlignment.at(j-1);
						wp2=(*columnPairLikelihood)[s2];
					}
					if (wp1>0 && wp2>0) {
						x=log(wp1)+log(wp2);

						// Regel S -> axFyb
						if(logStoaxFyb+x+gamma['F'][i+2][j-i-4]>gamma['S'][i][j-i]) {
							gamma['S'][i][j-i]=logStoaxFyb+x+gamma['F'][i+2][j-i-4];
							best_falt.setze('S',i,j-i,Faltungseintrag("((",'F',i+2,j-i-4,"))"));
						}
						// Regel L -> axFyb
						if(logLtoaxFyb+x+gamma['F'][i+2][j-i-4]>gamma['L'][i][j-i]) {
							gamma['L'][i][j-i]=logLtoaxFyb+x+gamma['F'][i+2][j-i-4];
							best_falt.setze('L',i,j-i,Faltungseintrag("((",'F',i+2,j-i-4,"))"));
						}
					}
				}
			}
		}
    }
    logwkeit=gamma['S'][0][n-1];
    faltung=best_falt.ausgabe();
    //return faltung=best_falt.ausgabe();
}

